"""Module to support working with Gaia epoch astrometry data.

Authors
-------

    Johannes Sahlmann



"""

from collections import OrderedDict
import logging
import os

from astropy.table import Table
from astropy.time import Time
import astropy.units as u
import numpy as np

from .pystrometry import OrbitSystem, convert_from_angular_to_linear, pjGet_m2, \
    AstrometricOrbitPlotter, MS_kg, MJ_kg, AstrometricPpmPlotter

def robust_scatter_estimate(data, verbose=True):
    """Return the Robust Scatter Estimate (RSE) es defined for Gaia.

    Parameters
    ----------
    data : ndarray
        Array of numbers.

    Returns
    -------

    References
    ----------

        @article{Lindegren:2012aa,
        Author = {{Lindegren}, L. and {Lammers}, U. and {Hobbs}, D. and {O'Mullane}, W. and {Bastian}, U. and {Hern{\'a}ndez}, J.},
        Journal = {\aap},
        Month = feb,
        Pages = {A78},
        Title = {{The astrometric core solution for the Gaia mission. Overview of models, algorithms, and software implementation}},
        Volume = 538,
        Year = 2012}

    """
    data_masked = np.ma.masked_array(data, np.isnan(data))
    data_cleaned = data_masked.data[data_masked.mask == False]
    if verbose:
        if len(data_masked) != len(data_cleaned):
            print('WARNING: data contains NaNs.')

    if not data_cleaned.size:
        return 0

    rse_factor = 0.390152
    rse = rse_factor * np.ptp(np.percentile(data_cleaned, [10, 90]))
    if np.isnan(rse):
        rse = 0.
    return rse



class GaiaIad(object):
    """Class for Gaia Epoch Astrometric Data."""

    def __init__(self, source_id, data_dir, epoch_data_suffix='_OBSERVATION_DATA_ALL.csv'):
        self.source_id = source_id
        self.data_dir = data_dir
        self.time_column = 'T-T0_yr'
        self.epoch_data_suffix = epoch_data_suffix

    def load_data(self, verbose=0, write_modified_iad=0, use_modified_epoch_data=False,
                  remove_outliers=True, filter_on_frame_uncertainty=False):

        self.epoch_data_file = os.path.join(self.data_dir, '{}{}'.format(self.source_id, self.epoch_data_suffix))
        self.epoch_data = Table.read(self.epoch_data_file)

        if 'direction_AL0_AC1' in self.epoch_data.colnames:
            if 0:
                self.xi = np.where(self.epoch_data['direction_AL0_AC1'] == 0)[0]
                self.yi = np.where(self.epoch_data['direction_AL0_AC1'] == 1)[0]
            else:
                remove_index = np.where(self.epoch_data['direction_AL0_AC1'] == 1)[0]
                self.epoch_data.remove_rows(remove_index)

        self.n_filtered_frames = 0
        self.n_original_frames = len(self.epoch_data)
        if filter_on_frame_uncertainty:
            rse_filter_threshold = 10
            rse = robust_scatter_estimate(self.epoch_data['errda_mas_obs'])
            remove_index = np.where(np.abs(self.epoch_data['errda_mas_obs']) > rse_filter_threshold*rse)[0]
            self.n_filtered_frames += len(remove_index)
            print('RSE cut removed {} frames (of {})'.format(self.n_filtered_frames, self.n_original_frames))
            self.epoch_data.remove_rows(remove_index)

        # remove focal plane transit averages used by GA code
        if 'isOriginal' in self.epoch_data.colnames:
            remove_index = np.where(self.epoch_data['isOriginal'] == 0)[0]
            self.epoch_data.remove_rows(remove_index)

        if remove_outliers:
            # remove_index = np.where(self.epoch_data['filterIntFlag'] == 2)[0]
            remove_index = np.where(self.epoch_data['filterIntFlag'] != 0)[0]
            self.epoch_data.remove_rows(remove_index)

        # sort by time
        self.epoch_data.sort(self.time_column)
        if verbose:
            print('Loaded IAD of {}'.format(self.source_id))
            print(self.epoch_data.info)

        # identify focal plane transits
        unique_transit_ids = np.unique(self.epoch_data['transitId'])
        self.n_fov_transit = len(unique_transit_ids)
        self.transit_id_name = 'OB'
        self.epoch_data[self.transit_id_name] = np.ones(len(self.epoch_data)).astype(int)
        for ob_number, transit_id in enumerate(unique_transit_ids):
            t_index = np.where(self.epoch_data['transitId'] == transit_id)[0]
            self.epoch_data[self.transit_id_name][t_index] = ob_number + 1

    def set_epoch_median_subtraction(self):
        """Subtract the median of an epoch from AL/AC measurements and add to table."""
        self.epoch_data['da_mas_obs_epoch_median_subtracted'] = np.zeros(len(self.epoch_data))
        for epoch_number in np.unique(self.epoch_data[self.transit_id_name]):
            t_index = np.where(self.epoch_data[self.transit_id_name] == epoch_number)[0]
            for direction_index in np.unique(self.epoch_data['direction_AL0_AC1'][t_index]):
                d_index = t_index[np.where(self.epoch_data['direction_AL0_AC1'][t_index] == direction_index)[0]]
                self.epoch_data['da_mas_obs_epoch_median_subtracted'][d_index] = self.epoch_data['da_mas_obs'][d_index] \
                                                                                 - np.median(self.epoch_data['da_mas_obs'][d_index])

    def astrometric_signal_to_noise(self, amplitude_mas):
        """Return astrometric signal to noise as defined in Sahlmann et al. 2011

        Parameters
        ----------
        amplitude_mas

        Returns
        -------

        """
        median_uncertainty_mas = np.median(self.epoch_data['errda_mas_obs'])
        n_ccd_transits = len(self.epoch_data)
        astrometric_snr = amplitude_mas * np.sqrt(n_ccd_transits)/median_uncertainty_mas

        return astrometric_snr


class GaiaValIad:
    """Class for Gaia Epoch Astrometric Data."""

    _transit_id_field = 'transitid'
    _fov_transit_id_field = 'OB'
    _obs_uncertainty_field = 'errda_mas_obs'
    scan_angle_definition = 'gaia'
    _time_field = 'relative_time_year'
    _mjd_field = 'MJD'
    time_column = 't-t_ref' # in years


    def __init__(self, source_id, epoch_data):
        self.source_id = source_id
        self.epoch_data = epoch_data
        self.n_filtered_frames = 0
        self.n_original_frames = len(self.epoch_data)

        self.epoch_df = self.epoch_data.to_pandas()

        # sort by time
        self.epoch_data.sort(self._time_field)
        self.set_fov_transit_id_field()

    # def set_fov_transit_id_field(self):
    #     # identify focal plane transits
    #     unique_transit_ids = self.epoch_data.to_pandas()[self._transit_id_field].unique() # not sorted
    #
    #     self.n_fov_transit = len(unique_transit_ids)
    #     self.epoch_data[self._fov_transit_id_field] = np.ones(len(self.epoch_data)).astype(int)
    #     for ob_number, transit_id in enumerate(unique_transit_ids):
    #         t_index = np.where(self.epoch_data[self._transit_id_field] == transit_id)[0]
    #         self.epoch_data[self._fov_transit_id_field][t_index] = ob_number + 1

    def set_reference_time(self, reference_time, **kwargs):
        """Set reference time.

        Parameters
        ----------
        reference_time : astropy.time
            Reference time used in calculations.

        """

        self.t_ref_mjd = reference_time.mjd
        self.reference_time = reference_time

        #  convert to absolute time in MJD
        if self._time_field == 'deltat':
            self.epoch_data['MJD'] = Time(self.epoch_data[self._time_field] * 365.25 + reference_time.jd, format='jd').mjd
#             self.epoch_data['MJD'] = Time(self.epoch_data[self._time_field] + reference_time.jd, format='jd').mjd
        elif self._time_field == 'elapsedNanoSecs':
            # public static final double REF_EPOCH_YR = 2010.0;
            reference_time = Time(2010.0, format='jyear')
            self.epoch_data['MJD'] = Time(self.epoch_data[self._time_field] * u.nanosecond.to(u.day) + reference_time.jd, format='jd').mjd
        elif self._time_field == 'relative_time_year':
            self.epoch_data['MJD'] = Time(self.epoch_data[self._time_field] * u.year.to(u.day) + reference_time.jd, format='jd').mjd
        self.epoch_df = self.epoch_data.to_pandas()

#         self.set_fov_transit_id_field()
#         self.sort_epochs_by_time()

        self.sort_epochs_by_time(time_column=self._mjd_field, **kwargs)
#         self.sort_epochs_by_time(time_column=self._time_field)

        self.set_fov_transit_id_field()

        self.epoch_data[self.time_column] = (self.epoch_data[self._mjd_field] - self.t_ref_mjd)/365.25
        self.epoch_df = self.epoch_data.to_pandas()

        self.verify_timing()


    def set_fov_transit_id_field(self, eliminate_transits_overlapping_in_time=True):
        """Identify focal plane transits and number them."""
        self.epoch_df[self._fov_transit_id_field] = self.epoch_df.groupby(self._transit_id_field,
                                                                          sort=False).ngroup()

        # check whether any of the transits overlap in time
        groups = self.epoch_df.groupby(self._fov_transit_id_field)
        min_max_time = groups.agg([min, max])[self._time_field]
        transits_overlapping_in_time = (min_max_time.shift(-1, fill_value=0)['min'] < min_max_time['max'])[:-1]
        if np.any(transits_overlapping_in_time):
            fov_transit_id_to_exclude = transits_overlapping_in_time[transits_overlapping_in_time == True].index
            logging.warning(f'These transit numbers are overlapping in time: {fov_transit_id_to_exclude}')
            if eliminate_transits_overlapping_in_time:
                logging.warning(f'Removing {len(fov_transit_id_to_exclude)} transits')
                self.epoch_df = self.epoch_df[self.epoch_df[self._fov_transit_id_field].isin(fov_transit_id_to_exclude) == False]

        self.epoch_data = Table.from_pandas(self.epoch_df)
        assert np.all(np.diff(self.epoch_data[self._fov_transit_id_field]) >= 0)

    def __str__(self):
        """Return string describing the instance."""
        return 'GaiaValIad for source_id {} with {} frames)'.format(self.source_id,
                                                                    len(self.epoch_data))
    def sort_epochs_by_time(self, time_column=None, remove_dubious_times=True):
        if time_column is None:
            time_column = self._mjd_field

        n_unique_times = self.epoch_df[time_column].nunique()
        n_rows_df = len(self.epoch_df)
        if n_unique_times != n_rows_df:
            logging.warning(f'Values in time field {time_column} are not unique. There are {n_rows_df-n_unique_times}/{n_rows_df} duplicates.')
            value_counts = self.epoch_df[time_column].value_counts()
            logging.debug('Rows with duplicated timestamps:')
            logging.debug('\n{}'.format(self.epoch_df[self.epoch_df[time_column].isin(value_counts[value_counts>1].index)].sort_values(time_column)[[time_column, self._time_field, self._transit_id_field, 'w', 'strip']]))
            if remove_dubious_times:
                dubious_times = self.epoch_df[self.epoch_df[time_column].isin(value_counts[value_counts>1].index)][time_column].values
                logging.debug(f'Removing data for {len(dubious_times)} times')
                self.epoch_df = self.epoch_df[self.epoch_df[time_column].isin(dubious_times) == False]
            self.epoch_data = Table.from_pandas(self.epoch_df)

        self.epoch_data.sort(time_column)
        self.epoch_df = self.epoch_data.to_pandas()
        assert np.all(self.epoch_data[time_column].data == self.epoch_df[time_column].values)


    @classmethod
    def from_dataframe(cls, df, source_id, df_source_id_field='source_id'):
        """Extract data from dataframe and return astropy table."""
        return cls(source_id,
                   Table.from_pandas(df.query('{}==@source_id'.format(df_source_id_field))))

    def filter_on_frame_uncertainty(self, filter='rse', absolute_filter_threshold=None, rse_filter_threshold=10):

        if filter=='rse':
            rse = robust_scatter_estimate(self.epoch_data[self._obs_uncertainty_field])
            remove_index = \
            np.where(np.abs(self.epoch_data[self._obs_uncertainty_field]) > rse_filter_threshold * rse)[
                0]
        elif filter=='simple':
            remove_index = np.where(np.abs(self.epoch_data[self._obs_uncertainty_field]) > absolute_filter_threshold)[0]

        self.n_filtered_frames += len(remove_index)
        logging.info('{} uncertainty cut removed {} frames (of {})'.format(filter, len(remove_index),
                                                         self.n_original_frames))
        self.epoch_data.remove_rows(remove_index)
        self.epoch_df = self.epoch_data.to_pandas()

    def filter_on_epoch_uncertainty(self, filter_threshold=2.5):

        epoch_signal_rms = self.epoch_data.group_by(self._fov_transit_id_field).groups.aggregate(np.std)['da_mas']
        indx = np.where(epoch_signal_rms > filter_threshold * np.median(epoch_signal_rms))[0]

        # indx = np.where(epoch_signal_rms > filter_threshold * robust_scatter_estimate(epoch_signal_rms))[0]

        epochs_to_exclude = np.unique(self.epoch_data.group_by(self._fov_transit_id_field).groups[indx][self._fov_transit_id_field].data)
        print(epochs_to_exclude)
        for epoch_to_exclude in epochs_to_exclude:
            remove_index = np.where(self.epoch_data[self._fov_transit_id_field] == epoch_to_exclude)[0]
            # if remove_index:
            self.n_filtered_frames += len(remove_index)
            logging.info('Filter on epoch uncertainty cut removed {} frames (of {})'.format(self.n_filtered_frames,
                                                             self.n_original_frames))
            self.epoch_data.remove_rows(remove_index)
        self.epoch_df = self.epoch_data.to_pandas()

    def filter_on_df_index(self, index_keep):
        """Keep only frames that are in index_keep."""
        n_current_frames = len(self.epoch_df)
        n_index_reject = n_current_frames - len(index_keep)
        self.n_filtered_frames += n_index_reject
        self.epoch_df = self.epoch_df.loc[index_keep]
        self.epoch_data = Table.from_pandas(self.epoch_df)
        logging.info('Filter on df index removed {} frames (of {})'.format(n_index_reject, n_current_frames))

    def filter_on_strip(self, strip_threshold=2):

        remove_index = np.where(self.epoch_data['strip'] <= strip_threshold)[0]
        self.n_filtered_frames += len(remove_index)
        logging.info('Filter on strip>{} removed {} frames (of {})'.format(strip_threshold, len(remove_index),
                                                         self.n_original_frames))
        self.epoch_data.remove_rows(remove_index)
        self.epoch_df = self.epoch_data.to_pandas()

    def to_miksga_csv(self, out_dir, source_id='all', return_df=False):
        """Export timeseries to csv."""
        csvdf = self.epoch_df.copy()
        rename_dict = {'source_id': 'sourceId', 't_min_t0_yr': 'T-T0_yr', 'transitid': 'transitId'}
        csvdf.rename(columns=rename_dict, inplace=True)
        columns = 'sourceId,T-T0_yr,cpsi_obs,spsi_obs,ppfact_obs,tcpsi_obs,tspsi_obs,da_mas_obs,errda_mas_obs,transitId'.split(',')
        if return_df:
            return csvdf[columns]
        else:
            outfile = os.path.join(out_dir, f'{source_id}_epoch_data_selection.csv')
            csvdf[columns].to_csv(outfile)
            logging.info(f'wrote {outfile}')

    def verify_timing(self):
#         epoch_df = self.epoch_data.to_pandas()
#         logging.debug(epoch_df[epoch_df[self._mjd_field].diff() <= 0])
#         logging.debug((epoch_df[self._mjd_field].diff() > 0).all())
#         assert (self.epoch_df[self._mjd_field].diff()[1:] >= 0).all()
#         assert (self.epoch_df[self._time_field].diff()[1:] > 0).all()
        assert (self.epoch_df[self._fov_transit_id_field].diff()[1:] >= 0).all()

        if np.any(np.diff(self.epoch_data[self._fov_transit_id_field]) < 0):
            logging.debug(f'_fov_transit_id_field {self._fov_transit_id_field} is not monotonically increasing with time. Fixing it.')


class GaiaLpcParquetIad(GaiaValIad):
    """Class for LPC Data from parquet dumper."""

    _time_field = 'elapsedNanoSecs'
    # _time_field = 'deltat' # in years
    # time_column = 't-t_ref' # in years
    _transit_id_field = 'transitId'
    # _fov_transit_id_field = 'OB'

    _obs_uncertainty_field = 'sigma_da_mas'
    # scan_angle_definition = 'gaia'
        
    
    def __init__(self, source_id, epoch_data):
        """Initialise object from pandas.DataFrame."""
#         super().__init__(source_id, epoch_data)
        self.source_id = source_id
        self.epoch_data = epoch_data
        # temporary unit fix
        self.epoch_data['alpha'] = self.epoch_data['ra']
        self.epoch_data['delta'] = self.epoch_data['dec']
        self.epoch_data['ra'] = np.rad2deg(self.epoch_data['alpha'])
        self.epoch_data['dec'] = np.rad2deg(self.epoch_data['delta'])

        self.n_filtered_frames = 0
        self.n_original_frames = len(self.epoch_data)
        
        self.epoch_df = self.epoch_data.to_pandas()
#         self.sort_epochs_by_time(time_column=self._time_field)
#         self.verify_missing_values()
        self.set_custom_fields()
        


    def verify_missing_values(self, remove_dubious_transits=True):    
        for key in ['w', 'obmt', 'theta']:
            if self.epoch_df[key].isnull().any():
                logging.warning(f'Some of the {key} data are masked.')
                if key == 'w':
                    logging.warning(f'Removing {len(self.epoch_df[key].isnull())}/{len(self.epoch_df)} rows where {key} data are masked.')
                    self.epoch_df = self.epoch_df[self.epoch_df[key].isnull() == False]
        
        # if len(self.epoch_df) == 0:
        #     raise ValueError('IAD do not contain valid AL measurements. (all null)')
        
        n_unique_times = self.epoch_df[self._time_field].nunique()
        n_rows_df = len(self.epoch_df)
        if n_unique_times != n_rows_df:
            logging.warning(f'Values in time field {self._time_field} are not unique. There are {n_rows_df-n_unique_times}/{n_rows_df} duplicates.')
            value_counts = self.epoch_df[self._time_field].value_counts()
            logging.debug('Rows with duplicated timestamps:')
            logging.debug('\n{}'.format(self.epoch_df[self.epoch_df[self._time_field].isin(value_counts[value_counts>1].index)][[self._time_field, self._transit_id_field, 'w', 'strip']]))
            # 1/0
#             logging.debug(f'epoch_data {self._time_field}:\n {epoch_df[self._time_field]}')
            if remove_dubious_transits:
                dubious_transit_ids = self.epoch_df[self.epoch_df[self._time_field].isin(value_counts[value_counts>1].index)][self._transit_id_field].values
                logging.debug(f'Removing data for {len(dubious_transit_ids)} transitIds')
                self.epoch_df = self.epoch_df[self.epoch_df[self._transit_id_field].isin(dubious_transit_ids) == False]
                
    
#         epoch_df.sort_values(self._time_field, inplace=True)
        self.epoch_data = Table.from_pandas(self.epoch_df)
    
    def set_custom_fields(self):
        
        self.epoch_data['spsi_obs'] = np.sin(self.epoch_data['theta'])
        self.epoch_data['cpsi_obs'] = np.cos(self.epoch_data['theta'])
        self.epoch_data['ppfact_obs'] = self.epoch_data['varpiFactorAl']
        self.epoch_data['da_mas'] = self.epoch_data['w']
        self.epoch_data['sigma_da_mas'] = self.epoch_data['wError']

        al_data = self.epoch_data['w'].data
        if (hasattr(al_data, 'mask')) and (np.all(al_data.mask)):
            logging.warning('IAD do not contain valid AL measurements. (all masked)')

        self.epoch_df = self.epoch_data.to_pandas()



            

class GaiaValIadHybrid(GaiaValIad):
    """Class for hybrid IAD that is produced internally in DU437 from a combination of StarObject and ObsConstStar data."""

    _time_field = 't_min_t0_yr'

    def set_reference_time(self, reference_time):
        """Set reference time.

        Parameters
        ----------
        reference_time : astropy.time
            Reference time used in calculations.

        """
        #  convert to absolute time in MJD
        self.epoch_data['MJD'] = Time(self.epoch_data[self._time_field] * 365.25 + reference_time.jd, format='jd').mjd

        self.epoch_data['t-t_ref'] = self.epoch_data[self._time_field]
        self.time_column = 't-t_ref'
        self.t_ref_mjd = reference_time.mjd
        self.reference_time = reference_time



def plot_individual_orbit(parameter_dict, iad, mapping_dr3id_to_starname=None,
                          plot_dir=os.path.expanduser('~'), m1_MS=1., degenerate_orbit=False):
    """Plot the orbit(s) of a single source due to orbiting companion(s).

    The code is prepared to handle multiple Keplerian solutions.

    Parameters
    ----------
    parameter_dict : dict
        Dictionary containing the orbit parameters
    iad : GaiaIad object
        Object holding the epoch astrometry data
    mapping_dr3id_to_starname : dict
        Mapping of source_id to common source name
    plot_dir : str
        Where to save the plot
    m1_MS : float
        Primary mass to use for companion mass estimation, under the assumption that the companion is dark.
    degenerate_orbit : bool
        Whether the orbit is 'degenerate'. This corresponds to the `other` set of geometric elements
        that produce the same astrometric orbit.

    Returns
    -------
    axp : AstrometricOrbitPlotter instance

    """
    model_parameters = OrderedDict()
    orbit_description = OrderedDict()

    # let's be tolerant in terms of variable naming
    parameter_dict = add_astrometric_parameter_equivalent_names(parameter_dict)


    # loop over every companion in system
    for planet_index in np.arange(parameter_dict.get('Nplanets', 1)):
        planet_number = planet_index + 1
        alpha_mas = parameter_dict['p{}_a1_mas'.format(planet_number)]
        absolute_parallax_mas = parameter_dict.get('parallax')
        a_m = convert_from_angular_to_linear(alpha_mas, absolute_parallax_mas)
        P_day = parameter_dict['p{}_period_day'.format(planet_number)]
        m2_kg = pjGet_m2(m1_MS * MS_kg, a_m, P_day)
        m2_MJ = m2_kg / MJ_kg
        m2_MS = m2_kg / MS_kg
        mass_function = alpha_mas**3 * (u.year.to(u.day))**2 / (P_day**2 * absolute_parallax_mas**3)

        # create dictionary that defines the OrbitSystem
        attribute_dict = {
            'offset_alphastar_mas': parameter_dict.get('alphaStarOffset_mas', 0),
            'offset_delta_mas': parameter_dict.get('deltaOffset_mas', 0),
            'RA_deg': parameter_dict['ra'],
            'DE_deg': parameter_dict['dec'],
            'absolute_plx_mas': parameter_dict.get('parallax'),
            'muRA_mas': parameter_dict['muAlphaStar_masPyr'],
            'muDE_mas': parameter_dict['muDelta_masPyr'],
            'P_day': parameter_dict['p{}_period_day'.format(planet_number)],
            'ecc': parameter_dict['p{}_ecc'.format(planet_number)],
            'omega_deg': parameter_dict['p{}_omega_deg'.format(planet_number)],
            'OMEGA_deg': parameter_dict['p{}_OMEGA_deg'.format(planet_number)],
            'i_deg': parameter_dict['p{}_incl_deg'.format(planet_number)],
            'a_mas': parameter_dict['p{}_a1_mas'.format(planet_number)],
            'Tp_day': iad.t_ref_mjd + parameter_dict['p{}_Tp_day-T0'.format(planet_number)],  # this is in MJD
            'm1_MS': m1_MS,
            'm2_MJ': m2_MJ,
            'Tref_MJD': iad.t_ref_mjd,
            'scan_angle_definition': iad.scan_angle_definition,
            'solution_type': parameter_dict.get('nss_solution_type', '')
        }
        # print(iad.scan_angle_definition)
        if degenerate_orbit:
            attribute_dict['omega_deg'] += 180.
            attribute_dict['OMEGA_deg'] += 180.

        if planet_index == 0:
            orbit = OrbitSystem(attribute_dict=attribute_dict)
            print(orbit)
            # set coeffMatrix in orbit object
            ppm_signal_mas = orbit.ppm(iad.epoch_data['MJD'], psi_deg=np.rad2deg(
                np.arctan2(iad.epoch_data['spsi_obs'], iad.epoch_data['cpsi_obs'])),
                                       offsetRA_mas=attribute_dict['offset_alphastar_mas'],
                                       offsetDE_mas=attribute_dict['offset_delta_mas'],
                                       externalParallaxFactors=iad.epoch_data['ppfact_obs'],
                                       verbose=True)

        model_parameters[planet_index] = attribute_dict

        if 'varpiError' in parameter_dict.keys():
            ppm_description = ''
            ppm_description += '$\\varpi={0[varpi]:2.{prec}f}\\pm{0[varpiError]:2.{prec}f}$ mas\n'.format(parameter_dict, prec=3)
            ppm_description += '$\mu_\\mathrm{{ra^\\star}}={0[pmra]:2.{prec}f}\\pm{0[pmraError]:2.{prec}f}$ mas\n'.format(parameter_dict, prec=3)
            ppm_description += '$\mu_\\mathrm{{dec}}={0[pmdec]:2.{prec}f}\\pm{0[pmdecError]:2.{prec}f}$ mas\n'.format(parameter_dict, prec=3)
        else:
            ppm_description = 'default'

        if ('sigma_p1_a1_mas' in parameter_dict.keys()) and (parameter_dict['Nplanets'] == 1):
            # temporary: only for single-companion solutions
            # orbit_descr = '$\\alpha={0[p1_a1_mas]:2.{prec}f}\\pm{0[sigma_p1_a1_mas]:2.{prec}f}$ mas (ratio={0[p1_a1_div_sigma_a1_mas]:2.1f})\n'.format(
            #     dict(parameter_dict), prec=3)
            # orbit_descr += '$P={0[p1_period_day]:2.{prec}f}\\pm{0[sigma_p1_period_day]:2.{prec}f}$ d\n'.format(
            #     dict(parameter_dict), prec=1)
            # orbit_descr += '$e={0[p1_ecc]:2.{prec}f}\\pm{0[sigma_p1_ecc]:2.{prec}f}$\n'.format(
            #     dict(parameter_dict), prec=3)
            # orbit_descr += '$i={0[p1_incl_deg]:2.{prec}f}$ deg\n'.format(dict(parameter_dict),
            #                                                              prec=2)
            # orbit_descr += '$\\omega={0[p1_omega_deg]:2.{prec}f}$ deg\n'.format(
            #     dict(parameter_dict), prec=2)
            # orbit_descr += '$\\Omega={0[p1_OMEGA_deg]:2.{prec}f}$ deg\n'.format(
            #     dict(parameter_dict), prec=2)
            orbit_descr = '$\\alpha={0[p1_a1_mas]:2.{prec}f}\\pm{0[sigma_p1_a1_mas]:2.{prec}f}$ mas (ratio={0[p1_a1_div_sigma_a1_mas]:2.1f})\n'.format(dict(parameter_dict), prec=3)
            orbit_descr += '$P={0[p1_period_day]:2.{prec}f}\\pm{0[sigma_p1_period_day]:2.{prec}f}$ d\n'.format(dict(parameter_dict), prec=1)
            orbit_descr += '$e={0[p1_ecc]:2.{prec}f}\\pm{0[sigma_p1_ecc]:2.{prec}f}$\n'.format(dict(parameter_dict), prec=3)
            orbit_descr += '$i={0[p1_incl_deg]:2.{prec}f}\\pm{0[sigma_p1_incl_deg]:2.{prec}f}$ deg\n'.format(dict(parameter_dict), prec=2)
            orbit_descr += '$\\omega={0[p1_omega_deg]:2.{prec}f}\\pm{0[sigma_p1_omega_deg]:2.{prec}f}$ deg\n'.format(dict(parameter_dict), prec=2)
            orbit_descr += '$\\Omega={0[p1_OMEGA_deg]:2.{prec}f}\\pm{0[sigma_p1_OMEGA_deg]:2.{prec}f}$ deg\n'.format(dict(parameter_dict), prec=2)
            orbit_descr += '$f(m)={0:2.{prec}f}$ Msun\n'.format(mass_function, prec=2)
            orbit_descr += '$M_1={0:2.{prec}f}$ Msun\n'.format(m1_MS, prec=2)
            orbit_descr += '$M_2={0:2.{prec}f}$ Msun\n'.format(m2_MS, prec=2)
            orbit_descr += '$M_2={0:2.{prec}f}$ Mjup\n'.format(m2_MJ, prec=2)

        else:
            orbit_descr = 'default'
        orbit_description[planet_index] = orbit_descr

    plot_dict = {}
    plot_dict['model_parameters'] = model_parameters
    plot_dict['linear_coefficients'] = {'matrix': orbit.coeffMatrix}
    if 'data_type' in parameter_dict.keys():
        plot_dict['data_type'] = parameter_dict['data_type']
    else:
        if hasattr(iad, 'xi'):
            plot_dict['data_type'] = 'gaia_2d'
        else:
            plot_dict['data_type'] = '1d'
    logging.debug(f"plot_dict['data_type'] = {plot_dict['data_type']}")

    plot_dict['scan_angle_definition'] = iad.scan_angle_definition

    for key in iad.epoch_data.colnames:
        if '_obs' in key:
            new_key = key.replace('_obs', '')
            if new_key == 'errda_mas':
                new_key = 'sigma_da_mas'
            iad.epoch_data[new_key] = iad.epoch_data[key]

    plot_dict['data'] = iad

    axp = AstrometricOrbitPlotter(plot_dict)

    n_curve = 1500
    timestamps_curve_2d = np.linspace(np.min(iad.epoch_data['MJD']), np.max(iad.epoch_data['MJD']),
                                      n_curve)
    axp.t_curve_MJD = timestamps_curve_2d

    if 'phot_g_mean_mag' in parameter_dict.keys():
        mag_str = '$G$={:2.1f}'.format(parameter_dict['phot_g_mean_mag'])
    else:
        mag_str = ''

    if mapping_dr3id_to_starname is not None:
        axp.title = 'Gaia DR3 {} ({} {})'.format(iad.source_id, mapping_dr3id_to_starname[source_id],
                                                mag_str)
        name_seed = 'DR3_{}_{}'.format(iad.source_id,
                                       mapping_dr3id_to_starname[iad.source_id].replace('/', '-'))
    else:
        # axp.title = 'Gaia DR3 {} ({})'.format(iad.source_id, mag_str)
        # name_seed = 'DR3_{}'.format(iad.source_id)
        axp.title = '{} ({})'.format(iad.source_id, mag_str)
        name_seed = '{}'.format(iad.source_id)

    if ('title' in parameter_dict.keys()) and (parameter_dict['title'] is not None):
        axp.title = parameter_dict['title']


    argument_dict = {'plot_dir'             : plot_dir, 'ppm_panel': True,
                     'frame_residual_panel' : True, 'orbit_only_panel': True,
                     'ppm_description'      : ppm_description,
                     'epoch_omc_description': 'default',
                     'orbit_description'    : orbit_description, 'arrow_offset_x': +100,
                     'arrow_offset_y'       : +100, 'name_seed': name_seed,
                     'scan_angle_definition': iad.scan_angle_definition}

    argument_dict['save_plot'] = True
    argument_dict['omc_panel'] = True
    argument_dict['orbit_only_panel'] = False
    argument_dict['make_condensed_summary_figure'] = False
    for key in ['make_xy_residual_figure', 'make_1d_overview_figure', 'orbit_only_panel', 'orbit_description',
                'epoch_omc_description', 'orbit_signal_description', 'frame_residual_panel', 'ppm_description']:

        if key in parameter_dict.keys():
            argument_dict[key] = parameter_dict[key]

        # argument_dict['make_xy_residual_figure'] = True
        # argument_dict['make_1d_overview_figure'] = False
    argument_dict['excess_noise'] = parameter_dict.get('excessNoise_mas', 0.)
    argument_dict['merit_function'] = parameter_dict.get('meritFunction', 0.)

    axp.argument_dict = argument_dict
    axp.plot(argument_dict=argument_dict)

    return axp

# small helper dictionary to become tolerant against different variable naming, e.g.
# internal DPAC naming versus Gaia archive naming
astrometric_parameter_name_equivalences = {'varpi': 'parallax',
                                           'varpi': 'plx_mas',
                                           'plx_mas': 'parallax',
                                           'plx_mas': 'absolute_plx_mas',
                                           'absolute_plx_mas': 'parallax',
                                           # '': 'parallax',
                                           'alpha0_deg': 'ra',
                                           'delta0_deg': 'dec',
                                           'alpha0_deg': 'RA_deg',
                                           'delta0_deg': 'DE_deg',
                                           'sourceId': 'source_id',
                                           'muAlphaStar_masPyr': 'pmra',
                                           'muDelta_masPyr': 'pmdec',
                                           'varpiError': 'parallax_error',
                                           'pmraError': 'pmra_error',
                                           'pmdecError': 'pmdec_error',
                                           'p1_ecc': 'eccentricity',
                                           'p1_Tp_day-T0': 't_periastron',
                                           'alphaStarOffset_mas': 'offset_alphastar_mas',
                                           'deltaOffset_mas': 'offset_delta_mas',
                                           }


def add_astrometric_parameter_equivalent_names(input_dictionary):
    """Modify input_dictionary by adding equivalent keys."""
    for key, equivalent_key in astrometric_parameter_name_equivalences.items():
        if (key not in input_dictionary.keys()) and (equivalent_key in input_dictionary.keys()):
            input_dictionary[key] = input_dictionary[equivalent_key]
        if (key in input_dictionary.keys()) and (equivalent_key not in input_dictionary.keys()):
            input_dictionary[equivalent_key] = input_dictionary[key]

    return input_dictionary


def plot_individual_ppm(parameter_dict, iad, plot_dir=os.path.expanduser('~')):
    """Make a 4-panel plot displaying the on-sky parallax+proper motion and residuals.

    Parameters
    ----------
    parameter_dict : dict
        Parameters of the astrometric model, e.g. parallax, proper motion
    iad :
        Epoch astrometry data for plotting.
    plot_dir : str
        Directory to store figure.

    """
    model_parameters = OrderedDict()
    orbit_description = OrderedDict()

    planet_index = 0

    # let's be tolerant in terms of variable naming
    parameter_dict = add_astrometric_parameter_equivalent_names(parameter_dict)

    attribute_dict = {
        'offset_alphastar_mas': parameter_dict.get('alphaStarOffset_mas', 0),
        'offset_delta_mas': parameter_dict.get('deltaOffset_mas', 0),
        'RA_deg': parameter_dict['ra'],
        'DE_deg': parameter_dict['dec'],
        'absolute_plx_mas': parameter_dict.get('parallax'),
        'muRA_mas': parameter_dict['pmra'],
        'muDE_mas': parameter_dict['pmdec'],
        'm2_MJ': 0.,
        'Tref_MJD': iad.t_ref_mjd,
        'scan_angle_definition': iad.scan_angle_definition,
    }
    attribute_dict['solution_type'] = parameter_dict.get('nss_solution_type', '')

    # logging.debug(attribute_dict)
    orbit = OrbitSystem(attribute_dict=attribute_dict)

    # set coeffMatrix in orbit object
    ppm_signal_mas = orbit.ppm(iad.epoch_data['MJD'].values, psi_deg=np.rad2deg(
        np.arctan2(iad.epoch_data['spsi_obs'], iad.epoch_data['cpsi_obs'])),
                               offsetRA_mas=attribute_dict['offset_alphastar_mas'], offsetDE_mas=attribute_dict['offset_delta_mas'],
                               externalParallaxFactors=iad.epoch_data['ppfact_obs'], verbose=True)

    model_parameters[planet_index] = attribute_dict

    orbit_descr = '{}  '.format('Scan-direction projected residuals added to model.')
    # orbit_descr += '\nScan angle spread = {:2.1f} deg'.format(np.ptp(np.rad2deg(iad.epoch_data['theta'])))

    # set defaults for uncertainties
    parameter_dict['parallax_error'] = parameter_dict.get('parallax_error', 0)
    parameter_dict['pmra_error'] = parameter_dict.get('pmra_error', 0)
    parameter_dict['pmdec_error'] = parameter_dict.get('pmdec_error', 0)

    ppm_description = ''
    ppm_description += '$\\varpi={0[parallax]:2.{prec}f}\\pm{0[parallax_error]:2.{prec}f}$ mas\n'.format(parameter_dict, prec=3)
    ppm_description += '$\mu_\\mathrm{{ra^\\star}}={0[pmra]:2.{prec}f}\\pm{0[pmra_error]:2.{prec}f}$ mas\n'.format(parameter_dict, prec=3)
    ppm_description += '$\mu_\\mathrm{{dec}}={0[pmdec]:2.{prec}f}\\pm{0[pmdec_error]:2.{prec}f}$ mas\n'.format(parameter_dict, prec=3)

    orbit_description[planet_index] = orbit_descr

    plot_dict = {}
    plot_dict['model_parameters'] = model_parameters
    plot_dict['linear_coefficients'] = {'matrix': orbit.coeffMatrix}

    if 'data_type' in parameter_dict.keys():
        plot_dict['data_type'] = parameter_dict['data_type']
    else:
        if hasattr(iad, 'xi'):
            plot_dict['data_type'] = 'gaia_2d'
        else:
            plot_dict['data_type'] = '1d'

    logging.debug(f"plot_dict['data_type'] = {plot_dict['data_type']}")
    plot_dict['scan_angle_definition'] = iad.scan_angle_definition

    # for key in iad.epoch_data.colnames:
    for key in iad.epoch_data.columns:
        if '_obs' in key:
            new_key = key.replace('_obs', '')
            if new_key == 'errda_mas':
                new_key = 'sigma_da_mas'
            iad.epoch_data[new_key] = iad.epoch_data[key]

    plot_dict['data'] = iad


    axp = AstrometricPpmPlotter(plot_dict)
    axp.set_residuals()

    n_curve = 1500
    timestamps_curve_2d = np.linspace(np.min(iad.epoch_data['MJD']), np.max(iad.epoch_data['MJD']), n_curve)
    axp.t_curve_MJD = timestamps_curve_2d

    if 'GMag' in parameter_dict.keys():
        mag_str = ' $G$={:2.1f}'.format(parameter_dict['GMag'])
    else:
        mag_str = ''

    source_id = np.int64(parameter_dict['sourceId'])
    # axp.title = 'Gaia DR4 {} ({}, {})'.format(source_id, 'LPC', mag_str)
    if ('title' in parameter_dict.keys()) and (parameter_dict['title'] is not None):
        axp.title = parameter_dict['title']
    else:
        axp.title = 'Gaia {} ({})'.format(source_id, mag_str)

    name_seed = parameter_dict.get('name_seed', 'Gaia_{}'.format(source_id))

    argument_dict = {'plot_dir': plot_dir, 'ppm_panel': True, 'frame_residual_panel': True,
             'ppm_description': ppm_description, 'epoch_omc_description': 'default',
             'orbit_description': orbit_description,
             'name_seed': name_seed, 'scan_angle_definition': iad.scan_angle_definition}

    argument_dict['save_plot'] = True
    argument_dict['omc_panel'] = True
    argument_dict['orbit_only_panel'] = False
    argument_dict['orbit_timeseries_panel'] = False
    argument_dict['make_condensed_summary_figure'] = False
    argument_dict['make_xy_residual_figure'] = False
    argument_dict['make_1d_overview_figure'] = True
    argument_dict['arrow_offset_x'] = parameter_dict['pmra']
    argument_dict['arrow_offset_x'] = parameter_dict['pmdec']
    
    argument_dict['excess_noise'] = parameter_dict.get('excessNoise', 0)
    argument_dict['merit_function'] = parameter_dict.get('meritFunction', -1)

    axp.plot(argument_dict=argument_dict)
    axp.argument_dict = argument_dict
    axp.parameter_dict = parameter_dict
    axp.attribute_dict = attribute_dict
    return axp



import numpy as np
import astropy.units as u

# constants according to BAS-045 Eq. 1
T_A_day = 1035. / 2 / np.sqrt(3)
T_B_day = 1035. / 2

year2day = u.year.to(u.day)

def offset_7p(accel, tau):
    """Offset function corresponding to acceleration term, see BAS-045 Eq. 2

    Note that accel is given in mas/year**2 whereas tau is in days. Therefore a conversion factor
    has to be applied at the end.

    Parameters
    ----------
    accel : float
        accel_ra or accel_dec in [mas/year**2]
    tau : float
        reference_time corrected time in days (\tilde tau = t - \tilde T)

    Returns
    -------

    """
    return 1. / 2 * accel * (tau ** 2 - T_A_day ** 2) / year2day**2


def offset_9p(deriv_accel, tau):
    """Offset function corresponding to acceleration term, see BAS-045 Eq. 2

    Parameters
    ----------
    deriv_accel : float
        deriv_accel_ra or deriv_accel_dec in [mas/year**3]
    tau : float
        reference_time corrected time in days (\tilde tau = t - \tilde T)

    Returns
    -------

    """
    return 1. / 6 * deriv_accel * (tau ** 2 - T_B_day ** 2) * tau / year2day**3



import errno
import os
import re
import shlex
import shutil
import subprocess
import sys
import textwrap

from distutils import log, ccompiler, sysconfig
from distutils.core import Extension
from distutils.ccompiler import get_default_compiler
from setuptools.command.build_ext import build_ext as SetuptoolsBuildExt
from setuptools.command import build_py

from ..utils import get_numpy_include_path, invalidate_caches, classproperty
from ..version_helpers import get_pkg_version_module


def should_build_with_cython(package, release=None):
    """Returns the previously used Cython version (or 'unknown' if not
    previously built) if Cython should be used to build extension modules from
    pyx files.  If the ``release`` parameter is not specified an attempt is
    made to determine the release flag from `astropy.version`.
    """

    try:
        version_module = __import__(package + '.cython_version',
                                    fromlist=['release', 'cython_version'])
    except ImportError:
        version_module = None

    if release is None and version_module is not None:
        try:
            release = version_module.release
        except AttributeError:
            pass

    try:
        cython_version = version_module.cython_version
    except AttributeError:
        cython_version = 'unknown'

    # Only build with Cython if, of course, Cython is installed, we're in a
    # development version (i.e. not release) or the Cython-generated source
    # files haven't been created yet (cython_version == 'unknown'). The latter
    # case can happen even when release is True if checking out a release tag
    # from the repository
    have_cython = False
    try:
        import Cython  # noqa
        have_cython = True
    except ImportError:
        pass

    if have_cython and (not release or cython_version == 'unknown'):
        return cython_version
    else:
        return False


_compiler_versions = {}


def get_compiler_version(compiler):
    if compiler in _compiler_versions:
        return _compiler_versions[compiler]

    # Different flags to try to get the compiler version
    # TODO: It might be worth making this configurable to support
    # arbitrary odd compilers; though all bets may be off in such
    # cases anyway
    flags = ['--version', '--Version', '-version', '-Version',
             '-v', '-V']

    def try_get_version(flag):
        process = subprocess.Popen(
            shlex.split(compiler, posix=('win' not in sys.platform)) + [flag],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        stdout, stderr = process.communicate()

        if process.returncode != 0:
            return 'unknown'

        output = stdout.strip().decode('latin-1')  # Safest bet
        if not output:
            # Some compilers return their version info on stderr
            output = stderr.strip().decode('latin-1')

        if not output:
            output = 'unknown'

        return output

    for flag in flags:
        version = try_get_version(flag)
        if version != 'unknown':
            break

    # Cache results to speed up future calls
    _compiler_versions[compiler] = version

    return version


# TODO: I think this can be reworked without having to create the class
# programmatically.
def generate_build_ext_command(packagename, release):
    """
    Creates a custom 'build_ext' command that allows for manipulating some of
    the C extension options at build time.  We use a function to build the
    class since the base class for build_ext may be different depending on
    certain build-time parameters (for example, we may use Cython's build_ext
    instead of the default version in distutils).

    Uses the default distutils.command.build_ext by default.
    """

    class build_ext(SetuptoolsBuildExt, object):
        package_name = packagename
        is_release = release
        _user_options = SetuptoolsBuildExt.user_options[:]
        _boolean_options = SetuptoolsBuildExt.boolean_options[:]
        _help_options = SetuptoolsBuildExt.help_options[:]

        force_rebuild = False

        _broken_compiler_mapping = [
            ('i686-apple-darwin[0-9]*-llvm-gcc-4.2', 'clang')
            ]

        # Warning: Spaghetti code ahead.
        # During setup.py, the setup_helpers module needs the ability to add
        # items to a command's user_options list.  At this stage we don't know
        # whether or not we can build with Cython, and so don't know for sure
        # what base class will be used for build_ext; nevertheless we want to
        # be able to provide a list to add options into.
        #
        # Later, once setup() has been called we should have all build
        # dependencies included via setup_requires available.  distutils needs
        # to be able to access the user_options as a *class* attribute before
        # the class has been initialized, but we do need to be able to
        # enumerate the options for the correct base class at that point

        @classproperty
        def user_options(cls):
            from distutils import core

            if core._setup_distribution is None:
                # We haven't gotten into setup() yet, and the Distribution has
                # not yet been initialized
                return cls._user_options

            return cls._final_class.user_options

        @classproperty
        def boolean_options(cls):
            # Similar to user_options above
            from distutils import core

            if core._setup_distribution is None:
                # We haven't gotten into setup() yet, and the Distribution has
                # not yet been initialized
                return cls._boolean_options

            return cls._final_class.boolean_options

        @classproperty
        def help_options(cls):
            # Similar to user_options above
            from distutils import core

            if core._setup_distribution is None:
                # We haven't gotten into setup() yet, and the Distribution has
                # not yet been initialized
                return cls._help_options

            return cls._final_class.help_options

        @classproperty(lazy=True)
        def _final_class(cls):
            """
            Late determination of what the build_ext base class should be,
            depending on whether or not Cython is available.
            """

            uses_cython = should_build_with_cython(cls.package_name,
                                                   cls.is_release)

            if uses_cython:
                # We need to decide late on whether or not to use Cython's
                # build_ext (since Cython may not be available earlier in the
                # setup.py if it was brought in via setup_requires)
                try:
                    from Cython.Distutils.old_build_ext import old_build_ext as base_cls
                except ImportError:
                    from Cython.Distutils import build_ext as base_cls
            else:
                base_cls = SetuptoolsBuildExt

            # Create and return an instance of a new class based on this class
            # using one of the above possible base classes
            def merge_options(attr):
                base = getattr(base_cls, attr)
                ours = getattr(cls, '_' + attr)

                all_base = set(opt[0] for opt in base)

                return base + [opt for opt in ours if opt[0] not in all_base]

            boolean_options = (base_cls.boolean_options +
                               [opt for opt in cls._boolean_options
                                if opt not in base_cls.boolean_options])

            members = dict(cls.__dict__)
            members.update({
                'user_options': merge_options('user_options'),
                'help_options': merge_options('help_options'),
                'boolean_options': boolean_options,
                'uses_cython': uses_cython,
            })

            # Update the base class for the original build_ext command
            build_ext.__bases__ = (base_cls, object)

            # Create a new class for the existing class, but now with the
            # appropriate base class depending on whether or not to use Cython.
            # Ensure that object is one of the bases to make a new-style class.
            return type(cls.__name__, (build_ext,), members)

        def __new__(cls, *args, **kwargs):
            # By the time the command is actually instantialized, the
            # Distribution instance for the build has been instantiated, which
            # means setup_requires has been processed--now we can determine
            # what base class we can use for the actual build, and return an
            # instance of a build_ext command that uses that base class (right
            # now the options being Cython.Distutils.build_ext, or the stock
            # setuptools build_ext)
            new_cls = super(build_ext, cls._final_class).__new__(cls._final_class)

            # Since the new cls is not a subclass of the original cls, we must
            # manually call its __init__
            new_cls.__init__(*args, **kwargs)
            return new_cls

        def finalize_options(self):
            # Add a copy of the _compiler.so module as well, but only if there
            # are in fact C modules to compile (otherwise there's no reason to
            # include a record of the compiler used)
            # Note, self.extensions may not be set yet, but
            # self.distribution.ext_modules is where any extension modules
            # passed to setup() can be found
            self._adjust_compiler()

            extensions = self.distribution.ext_modules
            if extensions:
                build_py = self.get_finalized_command('build_py')
                package_dir = build_py.get_package_dir(packagename)
                src_path = os.path.relpath(
                    os.path.join(os.path.dirname(__file__), 'src'))
                shutil.copy(os.path.join(src_path, 'compiler.c'),
                            os.path.join(package_dir, '_compiler.c'))
                ext = Extension(self.package_name + '._compiler',
                                [os.path.join(package_dir, '_compiler.c')])
                extensions.insert(0, ext)

            super(build_ext, self).finalize_options()

            # Generate
            if self.uses_cython:
                try:
                    from Cython import __version__ as cython_version
                except ImportError:
                    # This shouldn't happen if we made it this far
                    cython_version = None

                if (cython_version is not None and
                        cython_version != self.uses_cython):
                    self.force_rebuild = True
                    # Update the used cython version
                    self.uses_cython = cython_version

            # Regardless of the value of the '--force' option, force a rebuild
            # if the debug flag changed from the last build
            if self.force_rebuild:
                self.force = True

        def run(self):
            # For extensions that require 'numpy' in their include dirs,
            # replace 'numpy' with the actual paths
            np_include = get_numpy_include_path()
            for extension in self.extensions:
                if 'numpy' in extension.include_dirs:
                    idx = extension.include_dirs.index('numpy')
                    extension.include_dirs.insert(idx, np_include)
                    extension.include_dirs.remove('numpy')

                self._check_cython_sources(extension)

            super(build_ext, self).run()

            # Update cython_version.py if building with Cython
            try:
                cython_version = get_pkg_version_module(
                    packagename, fromlist=['cython_version'])[0]
            except (AttributeError, ImportError):
                cython_version = 'unknown'
            if self.uses_cython and self.uses_cython != cython_version:
                build_py = self.get_finalized_command('build_py')
                package_dir = build_py.get_package_dir(packagename)
                cython_py = os.path.join(package_dir, 'cython_version.py')
                with open(cython_py, 'w') as f:
                    f.write('# Generated file; do not modify\n')
                    f.write('cython_version = {0!r}\n'.format(self.uses_cython))

                if os.path.isdir(self.build_lib):
                    # The build/lib directory may not exist if the build_py
                    # command was not previously run, which may sometimes be
                    # the case
                    self.copy_file(cython_py,
                                   os.path.join(self.build_lib, cython_py),
                                   preserve_mode=False)

                invalidate_caches()

        def _adjust_compiler(self):
            """
            This function detects broken compilers and switches to another.  If
            the environment variable CC is explicitly set, or a compiler is
            specified on the commandline, no override is performed -- the
            purpose here is to only override a default compiler.

            The specific compilers with problems are:

                * The default compiler in XCode-4.2, llvm-gcc-4.2,
                  segfaults when compiling wcslib.

            The set of broken compilers can be updated by changing the
            compiler_mapping variable.  It is a list of 2-tuples where the
            first in the pair is a regular expression matching the version of
            the broken compiler, and the second is the compiler to change to.
            """
            if 'CC' in os.environ:
                # Check that CC is not set to llvm-gcc-4.2
                c_compiler = os.environ['CC']

                try:
                    version = get_compiler_version(c_compiler)
                except OSError:
                    msg = textwrap.dedent(
                        """
                        The C compiler set by the CC environment variable:

                            {compiler:s}

                        cannot be found or executed.
                        """.format(compiler=c_compiler))
                    log.warn(msg)
                    sys.exit(1)

                for broken, fixed in self._broken_compiler_mapping:
                    if re.match(broken, version):
                        msg = textwrap.dedent(
                            """Compiler specified by CC environment variable
                            ({compiler:s}:{version:s}) will fail to compile
                            {pkg:s}.

                            Please set CC={fixed:s} and try again.
                            You can do this, for example, by running:

                                CC={fixed:s} python setup.py <command>

                            where <command> is the command you ran.
                            """.format(compiler=c_compiler, version=version,
                                       pkg=self.package_name, fixed=fixed))
                        log.warn(msg)
                        sys.exit(1)

                # If C compiler is set via CC, and isn't broken, we are good to go. We
                # should definitely not try accessing the compiler specified by
                # ``sysconfig.get_config_var('CC')`` lower down, because this may fail
                # if the compiler used to compile Python is missing (and maybe this is
                # why the user is setting CC). For example, the official Python 2.7.3
                # MacOS X binary was compiled with gcc-4.2, which is no longer available
                # in XCode 4.
                return

            if self.compiler is not None:
                # At this point, self.compiler will be set only if a compiler
                # was specified in the command-line or via setup.cfg, in which
                # case we don't do anything
                return

            compiler_type = ccompiler.get_default_compiler()

            if compiler_type == 'unix':
                # We have to get the compiler this way, as this is the one that is
                # used if os.environ['CC'] is not set. It is actually read in from
                # the Python Makefile. Note that this is not necessarily the same
                # compiler as returned by ccompiler.new_compiler()
                c_compiler = sysconfig.get_config_var('CC')

                try:
                    version = get_compiler_version(c_compiler)
                except OSError:
                    msg = textwrap.dedent(
                        """
                        The C compiler used to compile Python {compiler:s}, and
                        which is normally used to compile C extensions, is not
                        available. You can explicitly specify which compiler to
                        use by setting the CC environment variable, for example:

                            CC=gcc python setup.py <command>

                        or if you are using MacOS X, you can try:

                            CC=clang python setup.py <command>
                        """.format(compiler=c_compiler))
                    log.warn(msg)
                    sys.exit(1)

                for broken, fixed in self._broken_compiler_mapping:
                    if re.match(broken, version):
                        os.environ['CC'] = fixed
                        break

        def _check_cython_sources(self, extension):
            """
            Where relevant, make sure that the .c files associated with .pyx
            modules are present (if building without Cython installed).
            """

            # Determine the compiler we'll be using
            if self.compiler is None:
                compiler = get_default_compiler()
            else:
                compiler = self.compiler

            # Replace .pyx with C-equivalents, unless c files are missing
            for jdx, src in enumerate(extension.sources):
                base, ext = os.path.splitext(src)
                pyxfn = base + '.pyx'
                cfn = base + '.c'
                cppfn = base + '.cpp'

                if not os.path.isfile(pyxfn):
                    continue

                if self.uses_cython:
                    extension.sources[jdx] = pyxfn
                else:
                    if os.path.isfile(cfn):
                        extension.sources[jdx] = cfn
                    elif os.path.isfile(cppfn):
                        extension.sources[jdx] = cppfn
                    else:
                        msg = (
                            'Could not find C/C++ file {0}.(c/cpp) for Cython '
                            'file {1} when building extension {2}. Cython '
                            'must be installed to build from a git '
                            'checkout.'.format(base, pyxfn, extension.name))
                        raise IOError(errno.ENOENT, msg, cfn)

                # Current versions of Cython use deprecated Numpy API features
                # the use of which produces a few warnings when compiling.
                # These additional flags should squelch those warnings.
                # TODO: Feel free to remove this if/when a Cython update
                # removes use of the deprecated Numpy API
                if compiler == 'unix':
                    extension.extra_compile_args.extend([
                        '-Wp,-w', '-Wno-unused-function'])

    return build_ext



# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2024-05-20 15:21:57 UTC
from __future__ import unicode_literals
import datetime

version = "3.0.2"
githash = "355a87d6e1b2a3cfc6ed5bf2da76bd6b8249533d"


major = 3
minor = 0
bugfix = 2

release = True
timestamp = datetime.datetime(2024, 5, 20, 15, 21, 57)
debug = False

astropy_helpers_version = ""

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"



import inspect
import sys
import re
import os
from warnings import warn

from sphinx.ext.autosummary.generate import find_autosummary_in_docstring


if sys.version_info[0] >= 3:
    def iteritems(dictionary):
        return dictionary.items()
else:
    def iteritems(dictionary):
        return dictionary.iteritems()

# We use \n instead of os.linesep because even on Windows, the generated files
# use \n as the newline character.
SPACE_NEWLINE = ' \n'
SINGLE_NEWLINE = '\n'
DOUBLE_NEWLINE = '\n\n'
TRIPLE_NEWLINE = '\n\n\n'


def cleanup_whitespace(text):
    """
    Make sure there are never more than two consecutive newlines, and that
    there are no trailing whitespaces.
    """

    # Get rid of overall leading/trailing whitespace
    text = text.strip() + '\n'

    # Get rid of trailing whitespace on each line
    while SPACE_NEWLINE in text:
        text = text.replace(SPACE_NEWLINE, SINGLE_NEWLINE)

    # Avoid too many consecutive newlines
    while TRIPLE_NEWLINE in text:
        text = text.replace(TRIPLE_NEWLINE, DOUBLE_NEWLINE)

    return text


def find_mod_objs(modname, onlylocals=False):
    """ Returns all the public attributes of a module referenced by name.

    .. note::
        The returned list *not* include subpackages or modules of
        `modname`,nor does it include private attributes (those that
        beginwith '_' or are not in `__all__`).

    Parameters
    ----------
    modname : str
        The name of the module to search.
    onlylocals : bool
        If True, only attributes that are either members of `modname` OR one of
        its modules or subpackages will be included.

    Returns
    -------
    localnames : list of str
        A list of the names of the attributes as they are named in the
        module `modname` .
    fqnames : list of str
        A list of the full qualified names of the attributes (e.g.,
        ``astropy.utils.misc.find_mod_objs``). For attributes that are
        simple variables, this is based on the local name, but for
        functions or classes it can be different if they are actually
        defined elsewhere and just referenced in `modname`.
    objs : list of objects
        A list of the actual attributes themselves (in the same order as
        the other arguments)

    """

    __import__(modname)
    mod = sys.modules[modname]

    if hasattr(mod, '__all__'):
        pkgitems = [(k, mod.__dict__[k]) for k in mod.__all__]
    else:
        pkgitems = [(k, mod.__dict__[k]) for k in dir(mod) if k[0] != '_']

    # filter out modules and pull the names and objs out
    ismodule = inspect.ismodule
    localnames = [k for k, v in pkgitems if not ismodule(v)]
    objs = [v for k, v in pkgitems if not ismodule(v)]

    # fully qualified names can be determined from the object's module
    fqnames = []
    for obj, lnm in zip(objs, localnames):
        if hasattr(obj, '__module__') and hasattr(obj, '__name__'):
            fqnames.append(obj.__module__ + '.' + obj.__name__)
        else:
            fqnames.append(modname + '.' + lnm)

    if onlylocals:
        valids = [fqn.startswith(modname) for fqn in fqnames]
        localnames = [e for i, e in enumerate(localnames) if valids[i]]
        fqnames = [e for i, e in enumerate(fqnames) if valids[i]]
        objs = [e for i, e in enumerate(objs) if valids[i]]

    return localnames, fqnames, objs


def find_autosummary_in_lines_for_automodsumm(lines, module=None, filename=None):
    """Find out what items appear in autosummary:: directives in the
    given lines.
    Returns a list of (name, toctree, template, inherited_members)
    where *name* is a name
    of an object and *toctree* the :toctree: path of the corresponding
    autosummary directive (relative to the root of the file name),
    *template* the value of the :template: option, and *inherited_members*
    is the value of the :inherited-members: option.
    *toctree*, *template*, and *inherited_members* are ``None`` if the
    directive does not have the corresponding options set.

    .. note::

       This is a slightly modified version of
       ``sphinx.ext.autosummary.generate.find_autosummary_in_lines``
       which recognizes the ``inherited-members`` option.
    """
    autosummary_re = re.compile(r'^(\s*)\.\.\s+autosummary::\s*')
    automodule_re = re.compile(
        r'^\s*\.\.\s+automodule::\s*([A-Za-z0-9_.]+)\s*$')
    module_re = re.compile(
        r'^\s*\.\.\s+(current)?module::\s*([a-zA-Z0-9_.]+)\s*$')
    autosummary_item_re = re.compile(r'^\s+(~?[_a-zA-Z][a-zA-Z0-9_.]*)\s*.*?')
    toctree_arg_re = re.compile(r'^\s+:toctree:\s*(.*?)\s*$')
    template_arg_re = re.compile(r'^\s+:template:\s*(.*?)\s*$')
    inherited_members_arg_re = re.compile(r'^\s+:inherited-members:\s*$')
    no_inherited_members_arg_re = re.compile(r'^\s+:no-inherited-members:\s*$')

    documented = []

    toctree = None
    template = None
    inherited_members = None
    current_module = module
    in_autosummary = False
    base_indent = ""

    for line in lines:
        if in_autosummary:
            m = toctree_arg_re.match(line)
            if m:
                toctree = m.group(1)
                if filename:
                    toctree = os.path.join(os.path.dirname(filename),
                                           toctree)
                continue

            m = template_arg_re.match(line)
            if m:
                template = m.group(1).strip()
                continue

            m = inherited_members_arg_re.match(line)
            if m:
                inherited_members = True
                continue

            m = no_inherited_members_arg_re.match(line)
            if m:
                inherited_members = False
                continue

            if line.strip().startswith(':'):
                warn(line)
                continue  # skip options

            m = autosummary_item_re.match(line)
            if m:
                name = m.group(1).strip()
                if name.startswith('~'):
                    name = name[1:]
                if current_module and \
                   not name.startswith(current_module + '.'):
                    name = "%s.%s" % (current_module, name)
                documented.append((name, toctree, template,
                                   inherited_members))
                continue

            if not line.strip() or line.startswith(base_indent + " "):
                continue

            in_autosummary = False

        m = autosummary_re.match(line)
        if m:
            in_autosummary = True
            base_indent = m.group(1)
            toctree = None
            template = None
            inherited_members = None
            continue

        m = automodule_re.search(line)
        if m:
            current_module = m.group(1).strip()
            # recurse into the automodule docstring
            documented.extend(find_autosummary_in_docstring(
                current_module, filename=filename))
            continue

        m = module_re.match(line)
        if m:
            current_module = m.group(2)
            continue

    return documented



"""
Functions to read and interpret results obtained with astromatic's scamp

"""
import copy
import os
from astropy.io.votable import parse
from astropy.io import fits
from astropy.wcs import WCS
import astropy.units as u
from astropy.time import Time
from astropy.coordinates import SkyCoord

from astropy.table import Table, Column
from astropy.table import vstack as tablevstack
import pylab as pl
import numpy as np

import pickle

try:
    import pystortion
except ImportError:
    print('WARNING: pystortion package not available!')

try:
    import aplpy
except ImportError:
    print('WARNING: aplpy package not available!')



class ScampXml(object):
    """
    Class for handling scamp.xml output file
    can be used as crossmatch table to identify stars across frames
    """
    def __init__(self, scamp_dir):
        self.table, self.source_file = getScampXml(scamp_dir)
        self.source_dir = scamp_dir
        # determine epoch number

        obs_date_mjd = Time(self.table['Observation_Date'].data, format='decimalyear').mjd
        diff_idx = np.where(np.diff(obs_date_mjd) > 0.1)[0]
        ob_numbers = np.ones(len(self.table)).astype(int)
        for j in diff_idx:
            ob_numbers[j + 1:] += 1
        self.table.add_column(Column(data=ob_numbers, name='EPOCH'))
        self.table.add_column(Column(data=obs_date_mjd, name='MJD'))

    # def plot(self, plot_dir):
    #     helpers.make_dir(plot_dir)
    #     helpers.plot_columns_simple(self.table, plot_dir)

    def info(self):
        self.target = str(np.unique(np.array(self.table['Image_Ident'])))
        self.n_frames = len(self.table)
        self.timespan_day = np.ptp(self.table['MJD'])
        self.n_epoch = len(np.unique(self.table['EPOCH']))
        self.average_pixel_scale = np.mean(self.table['Pixel_Scale'].data) * self.table['Pixel_Scale'].unit
        print('{}: target is {}, {} frames, {} epochs, {:2.1f} days timespan'.format(self.source_file, self.target, self.n_frames, self.n_epoch, self.timespan_day))
        # print('{}: Average pixel scale {:3.3f}'.format(self.source_file, self.average_pixel_scale.to(u.milliarcsecond)))

    def get_catalog_index(self, catalog_number=None, catalog_name=None):
        """
        return the table row index of a catalog
        :param catalog_number:
        :param catalog_name:
        :return:
        """
        if catalog_number is not None:
            catalog_index = self.table['Catalog_Number'].tolist().index(catalog_number)
        elif catalog_name is not None:
            catalog_index = self.table['Catalog_Name'].tolist().index((catalog_name.split('.')[0]+'.cat').encode())
        else:
            raise RuntimeError('Specify exactly one argument')
        return catalog_index

    def get_fits_filename(self, catalog_number):
        """
        return the name of the fits file used to create the catalog
        :param catalog_number:
        :return:
        """
        catalog_index = self.get_catalog_index(catalog_number=catalog_number)
        file_name = (self.table['Catalog_Name'][catalog_index]).decode().replace('.cat', '.fits')
        return file_name


class ScampFullCatalog(object):
    """
    Class for handling scamp full.cat output file
    http://scamp.readthedocs.io/en/latest/Output.html#full-catalogues
    """
    def __init__(self, scamp_dir=None, table=None):
        if table is not None:
            self.orig_table = table
            self.source_file = 'table_input'
            self.source_dir = None

        elif scamp_dir is not None:
            self.orig_table, self.source_file = read_scamp_cat(scamp_dir, 'full')
            self.source_dir = scamp_dir

        self.table = self.orig_table.copy()
        if 0 in self.orig_table['CATALOG_NUMBER'].data:
            # remove reference catalog
            self.table.remove_rows( np.where(self.orig_table['CATALOG_NUMBER']==0)[0] )

        self.unique_source_numbers = np.unique(self.table['SOURCE_NUMBER'])
        self.n_science_catalogs = len(np.unique(self.table['CATALOG_NUMBER']))

        # add column to identify fake stars used to stabilise astrometric reduction
        self.table['artificial'] = np.zeros(len(self.table)).astype(np.int)

    def info(self):
        """ Print basic info
        """
        print('{}: {} entries (including reference catalog)'.format(self.source_file, len(self.orig_table)))
        print('{}: {} entries (excluding reference catalog)'.format(self.source_file, len(self.table)))
        print('{}: {} catalogs (catalog 0 is the scamp reference catalog e.g. 2MASS)'.format(self.source_file, len(np.unique(self.orig_table['CATALOG_NUMBER']))))
        print('{}: {} science catalogs '.format(self.source_file, self.n_science_catalogs))
        print('{}: {} unique SOURCE_NUMBERs (excluding reference catalog)'.format(self.source_file, len(self.unique_source_numbers)))


    def complement_missing_catalogs(self, scamp_merged, threshold, verbose=False, target_number=None):
        """Fill in 'zero rows' for stars that have not been measured in all frames.

        if target_number is specified, no artifical measurements are added for the target

        This expands the original table.

        Returns
        -------

        """

        # identify sources that have not been measured in all frames/catalogs
        source_number_to_complement = np.unique(scamp_merged.table['SOURCE_NUMBER'][scamp_merged.table['NPOS_OK'] > threshold])
        if verbose:
            print('{}/{} sources have >{} scamp detections'.format(len(source_number_to_complement),
                                                               len(
                                                                   self.unique_source_numbers),
                                                               threshold))


        complemented_table = copy.deepcopy(self.table)

        all_catalog_numbers = np.unique(self.table['CATALOG_NUMBER'])
        if verbose:
            print('There are {} catalogs'.format(len(all_catalog_numbers)))

        # loop through sources
        for source_number in source_number_to_complement:
            if (target_number is not None) and (source_number==target_number):
                continue
            found_index = complemented_table['SOURCE_NUMBER'] == source_number
            found_in_catalog_numbers = complemented_table['CATALOG_NUMBER'][found_index]
            if verbose:
                print('Source {} found in {} catalogs'.format(source_number, len(found_in_catalog_numbers)))
            catalog_numbers_to_complement = np.setdiff1d(all_catalog_numbers,
                                                         found_in_catalog_numbers)
            if verbose:
                print('Have to complement in {} catalogs'.format(len(catalog_numbers_to_complement)))
            temp = copy.deepcopy(complemented_table[0:len(catalog_numbers_to_complement)])
            for colname in complemented_table.colnames:
                if colname == 'SOURCE_NUMBER':
                    temp[colname] = source_number
                elif colname == 'CATALOG_NUMBER':
                    temp[colname] = catalog_numbers_to_complement
                elif colname == 'artificial':
                    temp[colname] = 1
                # set X,Y pixel positions of artifical stars to something reasonable for reduced coordinates and basic functions
                elif colname in ['X_IMAGE', 'Y_IMAGE']:
                    temp[colname] = np.median(complemented_table[colname][found_index])
                else:
                    temp[colname] = np.zeros(len(temp)).astype(temp[colname].dtype)

            complemented_table = tablevstack((complemented_table, temp))

        if verbose:
            print('Number of rows before {} after {}'.format(len(self.table), len(complemented_table)))

        self.table = complemented_table

    def select_sources(self, catalog_number=None, chip_extension=None, source_number=None, ):
        """
        Select sources in a particular frame/catalog and chip extension
        :param catalog_number:
        :param chip_extension:
        :return:
        """

        table_index = range(len(self.table))

        if catalog_number is not None:
            table_index = np.intersect1d(table_index, np.where((self.table['CATALOG_NUMBER'] == catalog_number))[0])
        if chip_extension is not None:
            table_index = np.intersect1d(table_index, np.where((self.table['EXTENSION'] == chip_extension))[0])
        if source_number is not None:
            table_index = np.intersect1d(table_index, np.where((self.table['SOURCE_NUMBER'] == source_number))[0])

        return self.table[table_index]

    def remove_source(self, source_number):
        """
        Remove a source from catalog table

        :param source_number:
        :return:
        """
        self.table.remove_rows(np.where(self.table['SOURCE_NUMBER'] == source_number)[0])
        self.unique_source_numbers = np.unique(self.table['SOURCE_NUMBER'])

    def identify_source_number(self, RA_deg, DE_deg, catalog_number, chip_extension=1, verbose=False):
        """
        Return the source number that is closest to the (RA_deg, DE_deg position)

        :param RA_deg:
        :param DE_deg:
        :param catalog_number:
        :param chip_extension:
        :return:
        """

        # sources in catalog
        T = self.select_sources(catalog_number, chip_extension)
        cat = SkyCoord(ra=np.array(T['ALPHA_J2000']) * u.degree, dec=np.array(T['DELTA_J2000']) * u.degree)

        # coordinates of target source to identify
        target_cat = SkyCoord(ra=RA_deg * u.degree, dec=DE_deg * u.degree)

        # idx = np.where(target_cat.separation(cat).to(u.arcsec) < radius)[0]
        target_index = np.argmin(target_cat.separation(cat))
        target_source_number = T['SOURCE_NUMBER'][target_index]
        if verbose:
            print('Identified SOURCE_NUMBER {0} as closest {1} to expected target location'.format(target_source_number, np.min(
            target_cat.separation(cat)).to(u.arcsec)))

        return int(target_source_number)

    def plot_sources(self, scamp_xml, catalog_number, data_dir, chip_extension=1, save_plot=False, out_dir=None, RA_deg=None,
                     DE_deg=None, radius=50.*u.arcsec, tag='selected', label=True, external_cat=None):

        T = self.select_sources(catalog_number, chip_extension)
        # cat = SkyCoord(ra=np.array(T['RA_deg']) * u.degree, dec=np.array(T['DE_deg']) * u.degree)
        cat = SkyCoord(ra=np.array(T['ALPHA_J2000']) * u.degree, dec=np.array(T['DELTA_J2000']) * u.degree)


        # name of the individual frame FITS file being plotted
        file_name = scamp_xml.get_fits_filename(catalog_number)
        image_file = os.path.join(data_dir, file_name)

        header = fits.getheader(image_file)

        # catch case of OSIRIS where WCS is in second extension header, replace primary header by header of extension 2
        if 'DATE-OBS' not in header.keys():
            header = fits.getheader(image_file, ext=chip_extension)
            for key in header.keys():
                if 'PV' in key:
                    header.remove(key)

        # to align the plotted symbols with the actual image, use the WCS of the actual frame
        w = WCS(header=header)
        # coord_array = n.array([T['x'].data, T['y'].data]).T
        coord_array = np.array([T['X_IMAGE'].data, T['Y_IMAGE'].data]).T
        coord_array_world = w.wcs_pix2world(coord_array, 0)

        T['RA_deg'] = coord_array_world[:, 0]
        T['DE_deg'] = coord_array_world[:, 1]

        fig = pl.figure(figsize=(7, 7), facecolor='w', edgecolor='k'); pl.clf()
        try:
            gc = aplpy.FITSFigure(image_file, extension=chip_extension, figure=fig,
                                  North=True)  # ,convention='wells')#, dimensions=[1 ,0]);
        except:  # InvalidTransformError
            # for files reduced by Herve that contain the PV1_5 keywords
            data = fits.getdata(image_file, ext=chip_extension)
            hdu = fits.PrimaryHDU(data)
            hdu.header = header
            gc = aplpy.FITSFigure(hdu, extension=chip_extension, figure=fig,
                                  North=True)  # ,convention='wells')#, dimensions=[1 ,0]);

        # gc.show_grayscale(invert=False)  # , stretch='log', vmid=-1)#,vmid=-1)#, aspect = 1., pmax =90 )
        gc.show_grayscale(invert=False, stretch='log')  # , stretch='log', vmid=-1)#,vmid=-1)#, aspect = 1., pmax =90 )


        if RA_deg is not None:
            # target is defined
            gc.recenter(RA_deg, DE_deg, radius=radius.to(u.deg).value)
            gc.show_markers(RA_deg, DE_deg, marker='s', edgecolor='y', linewidth=3, s=30)
            target_cat = SkyCoord(ra=RA_deg * u.degree, dec=DE_deg * u.degree)
            idx = np.where(target_cat.separation(cat).to(u.arcsec) < radius)[0]
            target_source_number = self.identify_source_number(RA_deg, DE_deg, catalog_number, chip_extension, verbose=False)
        else:
            idx = range(len(cat))



        gc.show_markers(T['RA_deg'][idx] - 0 / 3600., T['DE_deg'][idx] - 0 / 3600., marker='o', edgecolor='g')
        if label:
            for j in idx:
                gc.add_label(T['RA_deg'][j] - 0 / 3600., T['DE_deg'][j] + 1 / 3600., T['SOURCE_NUMBER'][j])

        if external_cat is not None:
            gc.show_markers(external_cat.ra, external_cat.dec, marker='s', edgecolor='b', lw=1)


        pl.show()
        if save_plot:
            save_file = os.path.join(out_dir,file_name.replace('.fits', '_{}_sources.pdf'.format(tag)))
            gc.save(save_file, dpi=300)

    def correct_distortion(self, cal_pickle, calibration_catalog_number=None):
        """
        Corrects for distortion in every frame
        creates columns X_Idl and Y_Idl

        if calibration_catalog_number is specified, the full table is corrected for the distortion
        specified in the calibration_catalog. Otherwise every catalog is corrected for it's individual distortion

        :param cal_pickle:
        :return:
        """

        cals = pickle.load(open(cal_pickle, "rb"))  # cals is a dictionary

        self.table['X_Idl'] = np.zeros(len(self.table))
        self.table['Y_Idl'] = np.zeros(len(self.table))

        # this applies the transformation from the star catalog onto the Gaia frame
        evaluation_frame_number = 1

        if calibration_catalog_number is not None:
            cal_obs = cals[calibration_catalog_number]
            x_in = self.table['X_IMAGE']
            y_in = self.table['Y_IMAGE']
            self.table['X_Idl'], self.table['Y_Idl'] = \
                cal_obs.lazAC.apply_polynomial_transformation(evaluation_frame_number, x_in, y_in)
        else:
            catalog_numbers = self.table['CATALOG_NUMBER'].data
            for catalog_number in np.unique(catalog_numbers):
                table_index = np.where(self.table['CATALOG_NUMBER'] == catalog_number)[0]
                cal_obs = cals[catalog_number]

                x_in = self.table['X_IMAGE'][table_index]
                y_in = self.table['Y_IMAGE'][table_index]
                self.table['X_Idl'][table_index], self.table['Y_Idl'][
                    table_index] = cal_obs.lazAC.apply_polynomial_transformation(evaluation_frame_number, x_in, y_in)

    def suitable_reference_sources(self, selected_source_numbers, target_number, chip_extension):
        """
        return suitable reference stars (common in frames containing target, on same chip)
            TODO:
            - review reference star selection
            - check for number of astrometric instruments


        :param scamp_merged:
        :param selected_source_numbers:
        :param target_number:
        :param chip_extension:
        :return:
        """

        target_catalog_numbers = self.table['CATALOG_NUMBER'][self.table['SOURCE_NUMBER'] == target_number]
        target_catalog_numbers.sort()
        # target_frame_numbers = self.select_sources(source_number = target_number)

        print('Target identified in {} frames out of {}'.format(len(target_catalog_numbers), self.n_science_catalogs))


        source_numbers = selected_source_numbers
        good_source_numbers = []
        for i, num in enumerate(source_numbers):
            tmp = np.where(self.table['SOURCE_NUMBER'] == num)[0]
            if ((self.table['EXTENSION'][tmp[0]] == chip_extension) & (self.table['ASTR_INSTRUM'][tmp[0]] == 1) & (
                len(target_catalog_numbers) <= len(tmp))):  # select only stars that have equal or more detections than the target and are located in selected chip and catalogue 1
                # if set(self.table['CATALOG_NUMBER'][tmp]).issubset(set(target_catalog_numbers)):
                if set(target_catalog_numbers).issubset(set(self.table['CATALOG_NUMBER'][tmp])):
                # if all([j in tf['CATALOG_NUMBER'][tmp] for j in targetCatalogNumbers]):
                    good_source_numbers.append(num)
        print ('Selected {} reference stars as potential good reference stars'.format(len(good_source_numbers)));

        if target_number not in good_source_numbers:
            raise RuntimeError("Target not identified in all selected frames");

        return np.array(good_source_numbers).astype(np.int), target_catalog_numbers

    def get_astrometry(self, scamp_xml, red_dir, good_source_numbers, target_catalog_numbers, chip_extension, verbose=True, use_Idl=True):
        """
        Extract astrometry and auxiliary information

        TODO:
        fix assignment of 'sigma_x', 'sigma_y'

        :param scamp_xml:
        :param red_dir:
        :param good_source_numbers:
        :param target_catalog_numbers:
        :param chip_extension:
        :return:
        """
        n_stars = len(good_source_numbers)
        n_frames = len(target_catalog_numbers)

        p = np.zeros((n_frames, n_stars, 12))
        aux = np.zeros((n_frames, 16))

        tmplist = scamp_xml.table['Catalog_Number'].data.tolist()

        for i, catnum in enumerate(target_catalog_numbers):

            # find index in list of catalog numbers
            tmplist_idx = tmplist.index(catnum)

            # identify underlying FITS file
            frame_name = scamp_xml.table['Catalog_Name'][tmplist_idx].decode()
            frame = os.path.join(red_dir, frame_name.replace('.cat','.fits'))
            header = fits.getheader(frame)

            if verbose:
                print('{:03d}/{:03d}: filling catalogue with stars from {}.'.format(i+1,n_frames, frame))


            if header.get('ORIGIN') == None: # OSIRIS
                header = fits.getheader(frame, ext=2);
                aux[i, :] = [scamp_xml.table['Observation_Date'][tmplist_idx], 0.0, scamp_xml.table['AirMass'][tmplist_idx],
                             scamp_xml.table['Exposure_Time'][tmplist_idx], header['TAMBIENT'], header['PRESSURE'], header['HUMIDITY'],
                             header['WINDSPEE'], header['AZIMUTH'], header['ELEVAT'],
                             SkyCoord(dec=header['LATITUDE'], ra='00:00:00.000', unit='deg').dec.deg, header['DECDEG'],
                             header['RADEG'], header['MJD-OBS'], header['AIRMASS'], 0]

            elif header['INSTRUME'].strip() in ['GMOS-S', 'GMOS-N']:
                if header['INSTRUME'].strip() == 'GMOS-S':
                    latitude_string = '-30:14:26.700 '
                elif header['INSTRUME'].strip() == 'GMOS-N':
                    latitude_string = '19:49:25.7016'


                time_difference_day = np.abs(header['MJD-OBS'] - Time(scamp_xml.table['Observation_Date'][tmplist_idx], format='decimalyear').mjd)
                if time_difference_day > 1.:
                    1/0
                # scamp_xml.table['Observation_Date'][tmplist_idx]
                aux[i, :] = [Time(header['MJD-OBS'], format='mjd').decimalyear, catnum, header['AIRMASS'], header['EXPTIME'],
                             header['TAMBIENT'],
                             header['PRESSURE'], header['HUMIDITY'], header['WINDSPEE'], header['AZIMUTH'], header['ELEVATIO'],
                             SkyCoord(ra='00:00:00.000', dec=latitude_string, unit='deg').dec.deg, header['DEC'],
                             header['RA'],
                             header['MJD-OBS'], header['AIRMASS'], 0.0]

            else:  # HAWKI
                aux[i, :] = [scamp_xml.table['Observation_Date'][tmplist_idx], 0.0, header['HIERARCH ESO TEL AIRM START'],
                             header['HIERARCH ESO DET DIT'], header['HIERARCH ESO TEL AMBI TEMP'],
                             header['HIERARCH ESO TEL AMBI PRES START'], header['HIERARCH ESO TEL AMBI RHUM'],
                             header['HIERARCH ESO TEL AMBI WINDSP'], header['HIERARCH ESO TEL AZ'],
                             header['HIERARCH ESO TEL ALT'], header['HIERARCH ESO TEL GEOLAT'], header['DEC'], header['RA'],
                             header['MJD-OBS'], header['HIERARCH ESO TEL AIRM START'],
                             header['HIERARCH ESO ADA ABSROT START']]

            for j, IDN in enumerate(good_source_numbers):
                sidx = np.where((self.table['SOURCE_NUMBER'] == IDN) & (self.table['CATALOG_NUMBER'] == catnum))[0]
                #  flux, background level, X, Y, epochNr, seeing_X, seeing Y, IDN, chip
                if not use_Idl:
                    p[i, j, :] = [self.table['MAG'][sidx], self.table['FLAGS_EXTRACTION'][sidx], self.table['X_IMAGE'][sidx],
                                  self.table['Y_IMAGE'][sidx], i, self.table['ERRA_WORLD'][sidx], IDN, chip_extension,
                                  self.table['ERRA_IMAGE'][sidx], self.table['ERRB_IMAGE'][sidx]]
                else:
                    p[i, j, :] = [self.table['MAG'][sidx], self.table['FLAGS_EXTRACTION'][sidx], self.table['X_Idl'][sidx],
                                  self.table['Y_Idl'][sidx], i, self.table['ERRA_WORLD'][sidx], IDN, chip_extension,
                                  self.table['ERRA_IMAGE'][sidx], self.table['ERRB_IMAGE'][sidx], self.table['CATALOG_NUMBER'][sidx], self.table['artificial'][sidx]]

                # names of third dimension in p
                col_names = np.array(['MAG', 'FLAGS_EXTRACTION', 'x', 'y', 'index', 'sigma_world', 'id' , 'CHIP_EXTENSION', 'sigma_x', 'sigma_y', 'CATALOG_NUMBER', 'artificial'])

        # ATTENTION here: 'epoch_yr' stems from the scamp data 'Observation_Date' whereas 'MJD' comes from the file header 'MJD-OBS'
        aux = Table(aux, names=('scamp_epoch_yr', 'catalog_number', 'airmass', 'exptime', 'temperature', 'pressure' ,'rel_humidity', 'windspeed', 'tel_azimuth', 'tel_altitude', 'geo_latitude', 'dec' , 'ra', 'MJD', 'airmass2', 'abs_rotation'))
        if np.max(np.abs(aux['MJD'] - Time(aux['scamp_epoch_yr'], format='decimalyear').mjd)) > 1e-4:
            raise RuntimeError

        mp = pystortion.distortion.multiEpochAstrometry(p, col_names, aux)

        return mp


class ScampMergedCatalog(object):
    """
    Class for handling scamp merged.cat output file
    http://scamp.readthedocs.io/en/latest/Output.html#merged-catalogues
    """
    def __init__(self, scamp_dir):
        self.table, self.source_file = read_scamp_cat(scamp_dir, 'merged')
        self.source_dir = scamp_dir

    def info(self):
        print('{}: {} entries'.format(self.source_file, len(self.table)))

    def select_alignment_source_number(self, minimum_detections):
        """Return source numbers that will be used for alignment with Gaia/external catalog.

        Parameters
        ----------
        minimum_detections : int
            minimum number of scamp detections

        Returns
        -------
            numpy array

        """
        idx = np.where((self.table['FLAGS_SCAMP'] == 32) & (self.table['NPOS_OK'] > minimum_detections))[0]
        print('{}: {} sources have at least {} scamp detections and FLAGS_SCAMP == 32'.format(self.source_file, len(idx), minimum_detections))

        return np.array(self.table['SOURCE_NUMBER'][idx])


    def select_scamp_full(self, scamp_full, minimum_detections):
        """
        Return a ScampFullCatalog object that contains only sources that pass certain qualifiers

        :param minimum_detections:
        :return:
        """
        source_numbers = self.select_alignment_source_number(minimum_detections)
        idx_full = np.in1d(scamp_full.table['SOURCE_NUMBER'].data, source_numbers)
        scamp_full_select = ScampFullCatalog(table=scamp_full.table[idx_full])
        return scamp_full_select

    def remove_source(self, source_number):
        """
        Remove a source from catalog table

        :param source_number:
        :return:
        """
        self.table.remove_rows(np.where(self.table['SOURCE_NUMBER'] == source_number)[0])


    def source_numbers_within_radius(self, good_source_numbers, RA_deg, DE_deg, radius):
        """
        Return subset of good_source_numbers that lie within radius of Ra,Dec

        :param good_source_numbers:
        :param RA_deg:
        :param DE_deg:
        :param radius:
        :return:
        """
        source_number_list = self.table['SOURCE_NUMBER'].tolist()
        table_index = np.array([source_number_list.index(j) for j in good_source_numbers])
        cat = SkyCoord(ra=self.table['ALPHA_J2000'][table_index]*u.deg, dec=self.table['DELTA_J2000'][table_index]*u.deg)
        index = np.where(SkyCoord(ra=RA_deg*u.deg, dec=DE_deg*u.deg).separation(cat) < radius)[0]
        return good_source_numbers[index]


def getScampXml(targetDir, target_file=None):
    if target_file:
        pass;
    else:
        target_file = 'scamp.xml';
    votable = parse(os.path.join(targetDir, target_file));
    xmatchTable = votable.get_first_table().to_table();
    return xmatchTable, target_file


def read_scamp_cat(targetDir, tag, target_file=None):
    if target_file is None:
        target_file = '{}_1.cat'.format(tag)
    data = fits.getdata(os.path.join(targetDir, target_file), 2);
    tf = Table(data);
    return tf, target_file


def getScampFullCat(targetDir, target_file=None):
    if target_file:
        pass;
    else:
        target_file = 'full_1.cat';
    data = fits.getdata(os.path.join(targetDir, target_file), 2);
    tf = Table(data);
    return tf


def getScampMergedCat(targetDir, target_file=None):
    if target_file:
        pass;
    else:
        target_file = 'merged_1.cat';
    data = fits.getdata(os.path.join(targetDir, target_file), 2);
    tm = Table(data);
    return tm


def xfMakeAplpyPngForFrameInScampXml(tf, xmatchTable, dirlist, chip_extension, outDir, frameIdx=0, RA_deg=None,
                                     DE_deg=None):
    # number of entries in the full table
    scampCatLen = len(tf);
    print('number of entries in the full table %d' % scampCatLen)
    print('Number of individual SOURCE_NUMBER in full table: %d' % len(np.unique(tf['SOURCE_NUMBER'])))

    # name of the individual frame FITS file being plotted
    # frameName = xmatchTable['Catalog_Name'][frameIdx].replace('.cat', '.fits')
    frameName = (xmatchTable['Catalog_Name'][frameIdx]).decode().replace('.cat', '.fits')
    CatalogueNumber = xmatchTable['Catalog_Number'][frameIdx];

    # select stars that were detected in that frame
    starDict = [{"name": str(tf['SOURCE_NUMBER'][i]), "x": tf['X_IMAGE'][i], "y": tf['Y_IMAGE'][i]} for i in
                range(scampCatLen) if ((tf['EXTENSION'][i] == chip_extension) & (tf['ASTR_INSTRUM'][i] == 1) & (
        tf['CATALOG_NUMBER'][i] == CatalogueNumber))];

    T = Table(starDict)
    idx = [i for i, s in enumerate(dirlist) if frameName in s]
    image_file = dirlist[idx[0]]

    #     from astropy import wcs
    from astropy.wcs import WCS
    header = fits.getheader(image_file)

    # catch case of OSIRIS where WCS is in second extension header, replace primary header by header of extension 2
    if 'DATE-OBS' not in header.keys():
        header = fits.getheader(image_file, ext=chip_extension)
        for key in header.keys():
            if 'PV' in key:
                header.remove(key)

                #     1/0
                #     w = WCS(image_file)
    w = WCS(header=header)
    # coordArray = map(lambda x,y:[x,y],T['x'],T['y'])
    coordArray = np.array([T['x'].data, T['y'].data]).T
    # 1/0
    coordArray_world = w.wcs_pix2world(coordArray, 0)

    T['RA_deg'] = coordArray_world[:, 0]
    T['DE_deg'] = coordArray_world[:, 1]
    # find the frame image
    # print dataDir
    # print frameName
    # dirlist1=glob.glob('%s/%s' % (dataDir,frameName) );
    # if not dirlist1:
    #     dirlist1=glob.glob('%s**/**/%s' % (dataDir,frameName) );


    cat = SkyCoord(ra=np.array(T['RA_deg']) * u.degree, dec=np.array(T['DE_deg']) * u.degree)

    fig = pl.figure(figsize=(7, 7), facecolor='w', edgecolor='k');
    pl.clf();

    try:
        gc = aplpy.FITSFigure(image_file, extension=chip_extension, figure=fig,
                              North=True)  # ,convention='wells')#, dimensions=[1 ,0]);

    except:  # InvalidTransformError
        #
        #         # for files reduced by Herve that contain the PV1_5 keywords
        data = fits.getdata(image_file, ext=chip_extension)
        hdu = fits.PrimaryHDU(data)
        hdu.header = header
        gc = aplpy.FITSFigure(hdu, extension=chip_extension, figure=fig,
                              North=True)  # ,convention='wells')#, dimensions=[1 ,0]);

    gc.show_grayscale(
        invert=False)  # , stretch='log', vmid=-1)#,vmid=-1)#, aspect = pixScaleAC_mas/pixScaleAL_mas, pmax =90 )

    radius_as = 50.
    if RA_deg is not None:
        gc.recenter(RA_deg, DE_deg, radius=radius_as / 3600.)

    gc.show_markers(RA_deg, DE_deg, marker='s', edgecolor='y', linewidth=3, s=30)
    if 1 == 1:
        target_cat = SkyCoord(ra=RA_deg * u.degree, dec=DE_deg * u.degree)
        idx = np.where(target_cat.separation(cat).to(u.arcsec).value < radius_as)[0]
        gc.show_markers(T['RA_deg'][idx] - 0 / 3600., T['DE_deg'][idx] - 0 / 3600., marker='o', edgecolor='y')

        #         for j,lbl in  enumerate(T['name']):
        #             gc.add_label(T['RA_deg'][j]-0/3600.,T['DE_deg'][j]+1/3600.,lbl)
        for j in idx:
            gc.add_label(T['RA_deg'][j] - 0 / 3600., T['DE_deg'][j] + 1 / 3600., T['name'][j])

        target_index = np.argmin(target_cat.separation(cat))
        target_name = T['name'][target_index]
        print('Identified SOURCE_NUMBER {0} as closest {1} to expected target location'.format(target_name, np.min(
            target_cat.separation(cat)).to(u.arcsec)))

    saveFile = os.path.join(outDir,
                            xmatchTable.field('Catalog_Name')[frameIdx].decode().replace('.cat', '_SexStars.png'));
    gc.save(saveFile, dpi=300);
    pl.show()

    target_number = np.int(target_name)
    return target_number


def xf_make_frame_png_with_merged_scamp_catalog(tm, xmatchTable, dirlist, chipExtension, outDir, frameIdx=0,
                                                RA_deg=None, DE_deg=None):
    '''
    show target and reference star in FITS image with SOURCE_NUMBER
    fixed 2017-07-26 for WCS bug (yet not fully solved in xfMakeAplpyPngForFrameInScampXml)
    '''

    # number of entries in the full table
    scampCatLen = len(tm);
    print('number of entries in the merged table %d' % scampCatLen)
    print('Number of individual SOURCE_NUMBER in full table: %d' % len(np.unique(tm['SOURCE_NUMBER'])))

    # name of the individual frame FITS file being plotted
    frameName = xmatchTable['Catalog_Name'][frameIdx].replace('.cat', '.fits');
    CatalogueNumber = xmatchTable['Catalog_Number'][frameIdx];

    # select stars from merged catalog whether they were detected or not
    T = tm['SOURCE_NUMBER', 'ALPHA_J2000', 'DELTA_J2000']
    idx = [i for i, s in enumerate(dirlist) if frameName in s]
    image_file = dirlist[idx[0]]

    cat = SkyCoord(ra=np.array(T['ALPHA_J2000']) * u.degree, dec=np.array(T['DELTA_J2000']) * u.degree)

    fig = pl.figure(figsize=(7, 7), facecolor='w', edgecolor='k');
    pl.clf();

    gc = aplpy.FITSFigure(image_file, extension=chipExtension, figure=fig, North=True)
    gc.show_grayscale(
        invert=False)  # , stretch='log', vmid=-1)#,vmid=-1)#, aspect = pixScaleAC_mas/pixScaleAL_mas, pmax =90 )

    radius_as = 30.
    if RA_deg is not None:
        gc.recenter(RA_deg, DE_deg, radius=radius_as / 3600.)

    gc.show_markers(RA_deg, DE_deg, marker='s', edgecolor='y', linewidth=3, s=30)
    if 1 == 1:
        target_cat = SkyCoord(ra=RA_deg * u.degree, dec=DE_deg * u.degree)
        idx = np.where(target_cat.separation(cat).to(u.arcsec).value < radius_as)[0]
        gc.show_markers(T['ALPHA_J2000'][idx] - 0 / 3600., T['DELTA_J2000'][idx] - 0 / 3600., marker='o', edgecolor='y')

        for j in idx:
            gc.add_label(T['ALPHA_J2000'][j] - 0 / 3600., T['DELTA_J2000'][j] + 1 / 3600., T['SOURCE_NUMBER'][j])

        target_index = np.argmin(target_cat.separation(cat))
        print('Identified SOURCE_NUMBER {0} as closest {1} to expected target location'.format(
            T['SOURCE_NUMBER'][target_index], np.min(target_cat.separation(cat)).to(u.arcsec)))

    saveFile = os.path.join(outDir, xmatchTable.field('Catalog_Name')[frameIdx].replace('.cat', '_sources.png'));
    gc.save(saveFile, dpi=300);
    pl.show()


def xfDetermineTargetSourceNumber(tm, tmpIdx, RA_deg, DE_deg):  # , minimum_detections=0):
    #   gidx = np.where( tm['NPOS_OK'] > minimum_detections)[0];
    #     degDiffs = np.array([ np.sqrt( (RA_deg-tm['ALPHA_J2000'][i] )**2 + (DE_deg-tm['DELTA_J2000'][i] )**2 ) for i in tmpIdx ]);
    #     minDistIdx = np.argmin(degDiffs);
    #     target_number = tm['SOURCE_NUMBER'][tmpIdx[minDistIdx]];


    #     unique_SOURCE_NUMBER, unique_source_index = np.unique(tm['SOURCE_NUMBER'], return_index=True)

    print('number of entries in the merged table %d' % len(tm))
    print('Number of individual SOURCE_NUMBER in merged table: %d' % len(np.unique(tm['SOURCE_NUMBER'])))

    #     tm_unique = tm[unique_source_index]

    # catalog of detected stars in merged catalog
    cat = SkyCoord(ra=np.array(tm['ALPHA_J2000']) * u.degree, dec=np.array(tm['DELTA_J2000']) * u.degree)

    # expected position of target
    target_cat = SkyCoord(ra=RA_deg * u.degree, dec=DE_deg * u.degree)

    # separations
    distance_from_target = target_cat.separation(cat)

    # crossmatch radius
    radius = 10 * u.arcsec

    idx = np.where(distance_from_target < radius)[0]
    print('{0} stars in merged table are within {1} of the target'.format(len(idx), radius))
    for j in idx:
        print('{0}, distance {1}'.format(tm['SOURCE_NUMBER'][j], distance_from_target[j].to(u.arcsec)))

    target_index = np.argmin(distance_from_target.value)
    targetNumber = tm['SOURCE_NUMBER'][target_index];

    #     degDiffs = np.array([ np.sqrt( (RA_deg-tm['ALPHA_J2000'][i] )**2 + (DE_deg-tm['DELTA_J2000'][i] )**2 ) for i in tmpIdx ]);
    #     minDistIdx = np.argmin(degDiffs);
    #     target_number = tm['SOURCE_NUMBER'][tmpIdx[minDistIdx]];
    print('Target identified as reference star with SOURCE_NUMBER {:d} at {:3.4f}'.format(targetNumber, np.min(
        distance_from_target).to(u.arcsec)))

    #     print ("Target identified as reference star with SOURCE_NUMBER ##%d at %3.4f pix (%3.4f arcsec)" % (target_number,np.min(degDiffs)*3600./pixelscale,np.min(degDiffs)*3600.))
    if 0 == 1:
        pl.figure(1, figsize=(8, 5), facecolor='w', edgecolor='k');
        pl.clf();
        pl.plot(tmpIdx, degDiffs * 3600., 'ro');
        pl.ylim((0, 1000))
        plt.show()

    return targetNumber


def xfGetSuitableReferenceSources(tf, tm, tmpIdx, targetNumber, chipExtension):
    targetFrameNumbers = np.where(tf['SOURCE_NUMBER'] == targetNumber)[0];
    targetCatalogNumbers = tf['CATALOG_NUMBER'][targetFrameNumbers];
    targetCatalogNumbers.sort();
    print("Target identified in %d frames out of %d" % (len(targetFrameNumbers), len(np.unique(tf['CATALOG_NUMBER']))))

    #     TODO:
    #     - review reference star selection
    #     - check for number of astrometric instruments

    sourceNumbers = tm['SOURCE_NUMBER'][tmpIdx];
    goodSourceNumbers = [];
    for i, num in enumerate(sourceNumbers):
        tmp = np.where(tf['SOURCE_NUMBER'] == num)[0];
        if ((tf['EXTENSION'][tmp[0]] == chipExtension) & (tf['ASTR_INSTRUM'][tmp[0]] == 1) & (
            len(targetFrameNumbers) <= len(tmp))):  # select only stars in selected chip and catalogue 1
            if all([j in tf['CATALOG_NUMBER'][tmp] for j in targetCatalogNumbers]):
                goodSourceNumbers.append(num);
                #     pdb.set_trace()
    print ('Selected %d reference stars as potential good reference stars' % (len(goodSourceNumbers)));

    if targetNumber not in goodSourceNumbers:
        print ("target not identified in all selected frames: ERROR");
        1 / 0

    return np.array(goodSourceNumbers), targetFrameNumbers, targetCatalogNumbers


def xfGetReferenceSourcesInRopt(s_in, tm, RA_deg, DE_deg, R_opt_arcsec):
    # eliminate reference stars outside of circular area
    # s_out = [ j for j in s_in if (np.sqrt( (RA_deg-array(tm.field('ALPHA_J2000')[tm.field('SOURCE_NUMBER').tolist().index(j)]) )**2 + (DE_deg-array(tm.field('DELTA_J2000')[tm.field('SOURCE_NUMBER').tolist().index(j)]) )**2 ) < R_opt_arcsec/3600.) ];
    # degDiffs = np.array([ np.sqrt( (RA_deg-tm['ALPHA_J2000'][i] )**2 + (DE_deg-tm['DELTA_J2000'][i] )**2 ) for i in s_in ]);
    # idxs = np.array(np.where( degDiffs < R_opt_arcsec/3600. )[0]);
    # s_out = s_in[idxs]

    indexInTmOfSourceNumbers = np.array([tm['SOURCE_NUMBER'].tolist().index(j) for j in s_in]);
    degDiffs = np.sqrt((RA_deg - tm['ALPHA_J2000'][indexInTmOfSourceNumbers]) ** 2 + (
    DE_deg - tm['DELTA_J2000'][indexInTmOfSourceNumbers]) ** 2)
    idxs = np.array(np.where(degDiffs < R_opt_arcsec / 3600.)[0]);
    s_out = np.array(s_in[idxs]);
    # s_out = [ j for j in s_in if (np.sqrt( (RA_deg-tm['ALPHA_J2000'][] )**2 + (DE_deg-tm['DELTA_J2000'][tm['SOURCE_NUMBER'].tolist().index(j)] )**2 ) < R_opt_arcsec/3600.) ];
    #     print '%d remaining ref stars within R_opt' % len(s_out);
    return s_out


def xfGetInitialAstrometricData(xmatchTable, tf, dirlist, goodSourceNumbers, targetCatalogNumbers, chipExtension):
    Nstars = len(goodSourceNumbers);
    Nframes = len(targetCatalogNumbers);
    # Nframes  = len(targetFrameNumbers)

    p = numpy.zeros((Nframes, Nstars, 10))
    aux = numpy.zeros((Nframes, 16))
    tmplist = array(xmatchTable['Catalog_Number']).tolist();
    for i, catnum in enumerate(targetCatalogNumbers):
        tmplistIdx = tmplist.index(catnum)
        frameName = xmatchTable['Catalog_Name'][tmplistIdx].decode()
        frame = [dirlist[ii] for ii in range(len(dirlist)) if frameName.replace('.cat', '') in dirlist[ii]][0];
        #         print "%03d/%03d: filling catalogue with stars from %s" % (i+1,Nframes,frame);
        # head = fitsio.read_header(frame)           #
        head = fits.getheader(frame);
        # pdb.set_trace()
        if head.get('ORIGIN') == None:
            #         if not head.has_key('ORIGIN'): # for OSIRIS
            head = fits.getheader(frame, ext=2);
            # head['MJD-OBS']
            # print 'Inserting FITS keyword: MJD-OBS';
            # head.update('MJD-OBS', sla.slalib.sla_epj2d(xmatchTable.field('Observation_Date')[i]));
            # pdb.set_trace()
            aux[i, :] = [xmatchTable['Observation_Date'][i], 0.0, xmatchTable['AirMass'][i],
                         xmatchTable['Exposure_Time'][i], head['TAMBIENT'], head['PRESSURE'], head['HUMIDITY'],
                         head['WINDSPEE'], head['AZIMUTH'], head['ELEVAT'],
                         SkyCoord(ra=head['LATITUDE'], dec='00:00:00.000', unit='deg').ra.deg, head['DECDEG'],
                         head['RADEG'], head['MJD-OBS'], head['AIRMASS'], 0]

        elif head['INSTRUME'].strip() in ['GMOS-S', 'GMOS-N']:
            if head['INSTRUME'].strip() == 'GMOS-S':
                latitude_string = '-30:14:26.700 '
            elif head['INSTRUME'].strip() == 'GMOS-N':
                latitude_string = '19:49:25.7016'


                # removed head['HIERARCH ESO TEL GEOLAT']
            #             aux[i,:] = [  xmatchTable['Observation_Date'][i], 0.0 , head['AIRMASS'], head['EXPTIME'] ,head['TAMBIENT'], head['PRESSURE'], head['HUMIDITY'], head['WINDSPEE'], head['AZIMUTH'], head['ELEVATIO'], astCoords.dms2decimal('-30:14:26.700',':') , head['DEC'], head['RA'], head['MJD-OBS'], head['AIRMASS'], 0.0]
            aux[i, :] = [xmatchTable['Observation_Date'][i], 0.0, head['AIRMASS'], head['EXPTIME'], head['TAMBIENT'],
                         head['PRESSURE'], head['HUMIDITY'], head['WINDSPEE'], head['AZIMUTH'], head['ELEVATIO'],
                         SkyCoord(ra=latitude_string, dec='00:00:00.000', unit='deg').ra.deg, head['DEC'], head['RA'],
                         head['MJD-OBS'], head['AIRMASS'], 0.0]

        else:  # HAWKI
            aux[i, :] = [xmatchTable['Observation_Date'][i], 0.0, head['HIERARCH ESO TEL AIRM START'],
                         head['HIERARCH ESO DET DIT'], head['HIERARCH ESO TEL AMBI TEMP'],
                         head['HIERARCH ESO TEL AMBI PRES START'], head['HIERARCH ESO TEL AMBI RHUM'],
                         head['HIERARCH ESO TEL AMBI WINDSP'], head['HIERARCH ESO TEL AZ'],
                         head['HIERARCH ESO TEL ALT'], head['HIERARCH ESO TEL GEOLAT'], head['DEC'], head['RA'],
                         head['MJD-OBS'], head['HIERARCH ESO TEL AIRM START'], head['HIERARCH ESO ADA ABSROT START']]

        for j, IDN in enumerate(goodSourceNumbers):
            sidx = np.where((tf['SOURCE_NUMBER'] == IDN) & (tf['CATALOG_NUMBER'] == catnum))[0]
            #  flux, background level, X, Y, epochNr, seeing_X, seeing Y, IDN, chip
            # p[i,j,:] = [tf.field('MAG')[sidx], 0. , tf.field('X_IMAGE')[sidx] , tf.field('Y_IMAGE')[sidx] , i , 1 , IDN , chipExtension, tf['ERRA_IMAGE'][sidx] , tf['ERRB_IMAGE'][sidx]];
            p[i, j, :] = [tf.field('MAG')[sidx], tf.field('FLAGS_EXTRACTION')[sidx], tf.field('X_IMAGE')[sidx],
                          tf.field('Y_IMAGE')[sidx], i, tf.field('ERRA_WORLD')[sidx], IDN, chipExtension,
                          tf['ERRA_IMAGE'][sidx], tf['ERRB_IMAGE'][sidx]];
            # p[i,j,:] = [tf.field('MAG')[sidx], 0. , tf.field('X_IMAGE')[sidx] , tf.field('Y_IMAGE')[sidx] , i , int(frame.split('/')[6][2:]) , IDN , chipExtension, tf.field('ERRA_IMAGE')[sidx] , tf.field('ERRB_IMAGE')[sidx]];
            #         p[i,target_index,:] = [flx, bck , params[1] , params[2] , params[3] , params[4] , IDN , chip];

    return p, aux



from __future__ import print_function

import inspect
import os
import pkgutil
import re
import shutil
import subprocess
import sys
import textwrap
import warnings

from distutils import log
from distutils.cmd import DistutilsOptionError

import sphinx
from sphinx.setup_command import BuildDoc as SphinxBuildDoc

from ..utils import minversion, AstropyDeprecationWarning


class AstropyBuildDocs(SphinxBuildDoc):
    """
    A version of the ``build_docs`` command that uses the version of Astropy
    that is built by the setup ``build`` command, rather than whatever is
    installed on the system.  To build docs against the installed version, run
    ``make html`` in the ``astropy/docs`` directory.

    This also automatically creates the docs/_static directories--this is
    needed because GitHub won't create the _static dir because it has no
    tracked files.
    """

    description = 'Build Sphinx documentation for Astropy environment'
    user_options = SphinxBuildDoc.user_options[:]
    user_options.append(
        ('warnings-returncode', 'w',
         'Parses the sphinx output and sets the return code to 1 if there '
         'are any warnings. Note that this will cause the sphinx log to '
         'only update when it completes, rather than continuously as is '
         'normally the case.'))
    user_options.append(
        ('clean-docs', 'l',
         'Completely clean previous builds, including '
         'automodapi-generated files before building new ones'))
    user_options.append(
        ('no-intersphinx', 'n',
         'Skip intersphinx, even if conf.py says to use it'))
    user_options.append(
        ('open-docs-in-browser', 'o',
         'Open the docs in a browser (using the webbrowser module) if the '
         'build finishes successfully.'))

    boolean_options = SphinxBuildDoc.boolean_options[:]
    boolean_options.append('warnings-returncode')
    boolean_options.append('clean-docs')
    boolean_options.append('no-intersphinx')
    boolean_options.append('open-docs-in-browser')

    _self_iden_rex = re.compile(r"self\.([^\d\W][\w]+)", re.UNICODE)

    def initialize_options(self):
        SphinxBuildDoc.initialize_options(self)
        self.clean_docs = False
        self.no_intersphinx = False
        self.open_docs_in_browser = False
        self.warnings_returncode = False

    def finalize_options(self):

        SphinxBuildDoc.finalize_options(self)

        # Clear out previous sphinx builds, if requested
        if self.clean_docs:
            dirstorm = [os.path.join(self.source_dir, 'api'),
                        os.path.join(self.source_dir, 'generated')]
            if self.build_dir is None:
                dirstorm.append('docs/_build')
            else:
                dirstorm.append(self.build_dir)

            for d in dirstorm:
                if os.path.isdir(d):
                    log.info('Cleaning directory ' + d)
                    shutil.rmtree(d)
                else:
                    log.info('Not cleaning directory ' + d + ' because '
                             'not present or not a directory')

    def run(self):
        # TODO: Break this method up into a few more subroutines and
        # document them better
        import webbrowser

        from urllib.request import pathname2url

        # This is used at the very end of `run` to decide if sys.exit should
        # be called. If it's None, it won't be.
        retcode = None

        # If possible, create the _static dir
        if self.build_dir is not None:
            # the _static dir should be in the same place as the _build dir
            # for Astropy
            basedir, subdir = os.path.split(self.build_dir)
            if subdir == '':  # the path has a trailing /...
                basedir, subdir = os.path.split(basedir)
            staticdir = os.path.join(basedir, '_static')
            if os.path.isfile(staticdir):
                raise DistutilsOptionError(
                    'Attempted to build_docs in a location where' +
                    staticdir + 'is a file.  Must be a directory.')
            self.mkpath(staticdir)

        # Now make sure Astropy is built and determine where it was built
        build_cmd = self.reinitialize_command('build')
        build_cmd.inplace = 0
        self.run_command('build')
        build_cmd = self.get_finalized_command('build')
        build_cmd_path = os.path.abspath(build_cmd.build_lib)

        ah_importer = pkgutil.get_importer('astropy_helpers')
        ah_path = os.path.abspath(ah_importer.path)

        # Now generate the source for and spawn a new process that runs the
        # command.  This is needed to get the correct imports for the built
        # version
        runlines, runlineno = inspect.getsourcelines(SphinxBuildDoc.run)
        subproccode = textwrap.dedent("""
            from sphinx.setup_command import *

            os.chdir({srcdir!r})
            sys.path.insert(0, {build_cmd_path!r})
            sys.path.insert(0, {ah_path!r})

        """).format(build_cmd_path=build_cmd_path, ah_path=ah_path,
                    srcdir=self.source_dir)
        # runlines[1:] removes 'def run(self)' on the first line
        subproccode += textwrap.dedent(''.join(runlines[1:]))

        # All "self.foo" in the subprocess code needs to be replaced by the
        # values taken from the current self in *this* process
        subproccode = self._self_iden_rex.split(subproccode)
        for i in range(1, len(subproccode), 2):
            iden = subproccode[i]
            val = getattr(self, iden)
            if iden.endswith('_dir'):
                # Directories should be absolute, because the `chdir` call
                # in the new process moves to a different directory
                subproccode[i] = repr(os.path.abspath(val))
            else:
                subproccode[i] = repr(val)
        subproccode = ''.join(subproccode)

        optcode = textwrap.dedent("""

        class Namespace(object): pass
        self = Namespace()
        self.pdb = {pdb!r}
        self.verbosity = {verbosity!r}
        self.traceback = {traceback!r}

        """).format(pdb=getattr(self, 'pdb', False),
                    verbosity=getattr(self, 'verbosity', 0),
                    traceback=getattr(self, 'traceback', False))

        subproccode = optcode + subproccode

        # This is a quick gross hack, but it ensures that the code grabbed from
        # SphinxBuildDoc.run will work in Python 2 if it uses the print
        # function
        if minversion(sphinx, '1.3'):
            subproccode = 'from __future__ import print_function' + subproccode

        if self.no_intersphinx:
            # the confoverrides variable in sphinx.setup_command.BuildDoc can
            # be used to override the conf.py ... but this could well break
            # if future versions of sphinx change the internals of BuildDoc,
            # so remain vigilant!
            subproccode = subproccode.replace(
                'confoverrides = {}',
                'confoverrides = {\'intersphinx_mapping\':{}}')

        log.debug('Starting subprocess of {0} with python code:\n{1}\n'
                  '[CODE END])'.format(sys.executable, subproccode))

        # To return the number of warnings, we need to capture stdout. This
        # prevents a continuous updating at the terminal, but there's no
        # apparent way around this.
        if self.warnings_returncode:
            proc = subprocess.Popen([sys.executable, '-c', subproccode],
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT)

            retcode = 1
            with proc.stdout:
                for line in iter(proc.stdout.readline, b''):
                    line = line.strip(b'\r\n')
                    print(line.decode('utf-8'))
                    if 'build succeeded.' == line.decode('utf-8'):
                        retcode = 0

            # Poll to set proc.retcode
            proc.wait()

            if retcode != 0:
                if os.environ.get('TRAVIS', None) == 'true':
                    # this means we are in the travis build, so customize
                    # the message appropriately.
                    msg = ('The build_docs travis build FAILED '
                           'because sphinx issued documentation '
                           'warnings (scroll up to see the warnings).')
                else:  # standard failure message
                    msg = ('build_docs returning a non-zero exit '
                           'code because sphinx issued documentation '
                           'warnings.')
                log.warn(msg)

        else:
            proc = subprocess.Popen([sys.executable], stdin=subprocess.PIPE)
            proc.communicate(subproccode.encode('utf-8'))

        if proc.returncode == 0:
            if self.open_docs_in_browser:
                if self.builder == 'html':
                    absdir = os.path.abspath(self.builder_target_dir)
                    index_path = os.path.join(absdir, 'index.html')
                    fileurl = 'file://' + pathname2url(index_path)
                    webbrowser.open(fileurl)
                else:
                    log.warn('open-docs-in-browser option was given, but '
                             'the builder is not html! Ignoring.')
        else:
            log.warn('Sphinx Documentation subprocess failed with return '
                     'code ' + str(proc.returncode))
            retcode = proc.returncode

        if retcode is not None:
            # this is potentially dangerous in that there might be something
            # after the call to `setup` in `setup.py`, and exiting here will
            # prevent that from running.  But there's no other apparent way
            # to signal what the return code should be.
            sys.exit(retcode)


class AstropyBuildSphinx(AstropyBuildDocs): # pragma: no cover
    description = 'deprecated alias to the build_docs command'

    def run(self):
        warnings.warn(
            'The "build_sphinx" command is now deprecated. Use'
            '"build_docs" instead.', AstropyDeprecationWarning)
        AstropyBuildDocs.run(self)



try:
    from .version import version as __version__
    from .version import githash as __githash__
except ImportError:
    __version__ = ''
    __githash__ = ''


# If we've made it as far as importing astropy_helpers, we don't need
# ah_bootstrap in sys.modules anymore.  Getting rid of it is actually necessary
# if the package we're installing has a setup_requires of another package that
# uses astropy_helpers (and possibly a different version at that)
# See https://github.com/astropy/astropy/issues/3541
import sys
if 'ah_bootstrap' in sys.modules:
    del sys.modules['ah_bootstrap']


# Note, this is repeated from ah_bootstrap.py, but is here too in case this
# astropy-helpers was upgraded to from an older version that did not have this
# check in its ah_bootstrap.
# matplotlib can cause problems if it is imported from within a call of
# run_setup(), because in some circumstances it will try to write to the user's
# home directory, resulting in a SandboxViolation.  See
# https://github.com/matplotlib/matplotlib/pull/4165
# Making sure matplotlib, if it is available, is imported early in the setup
# process can mitigate this (note importing matplotlib.pyplot has the same
# issue)
try:
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot
except:
    # Ignore if this fails for *any* reason*
    pass


import os
# Ensure that all module-level code in astropy or other packages know that
# we're in setup mode:
if ('__main__' in sys.modules and
        hasattr(sys.modules['__main__'], '__file__')):
    filename = os.path.basename(sys.modules['__main__'].__file__)

    if filename.rstrip('co') == 'setup.py':
        import builtins
        builtins._ASTROPY_SETUP_ = True

    del filename



"""Module to support MCMC analyses of astrometric orbits.

Authors
-------
    - Johannes Sahlmann

"""

import numpy as np

def encode_eccentricity_omega(eccentricity, omega_deg):
    """Convert to alternative parameter set ot mitigate correlations and allow ecc=0.

    Parameters
    ----------
    eccentricity : float
    omega_deg : float
        Angle in degrees

    Returns
    -------
    x1, x2 : tuple of floats
        sqrt(ecc) * sin(omega), sqrt(ecc) * cos(omega)

    """
    omega_rad = np.deg2rad(omega_deg)
    x1 = np.sqrt(eccentricity) * np.sin(omega_rad)
    x2 = np.sqrt(eccentricity) * np.cos(omega_rad)

    return x1, x2


def decode_eccentricity_omega(x1, x2):
    """Convert back to original parameter set.

    Parameters
    ----------
    x1 : float
        sqrt(ecc) * sin(omega)
    x2: float
        sqrt(ecc) * cos(omega)

    Returns
    -------
    eccentricity : float
        Orbital eccentricity
    omega_deg : float
        omega in degrees in the range(-180, 180]

    """
    eccentricity = x1**2 + x2**2
    omega_deg = np.rad2deg(np.arctan2(x1, x2))

    return eccentricity, omega_deg


"""Some query helpers.

Authors
-------
    Johannes Sahlmann

"""
import logging
import os
from astropy.table import Table
from astroquery.gaia import Gaia
import pandas as pd


def get_gaiadr_data(analysis_dataset_name, data_dir, source_id_array=None, gaia_data_release='gaiadr3',
                    overwrite_query=False, gaia_table_name='gaia_source', shared_user_name=None):
    """Query a Gaia archive table by source_id. Only data corresponding to source_id_array are returned.

    Parameters
    ----------
    analysis_dataset_name
    data_dir
    source_id_array
    gaia_data_release
    overwrite_query
    gaia_table_name

    Returns
    -------

    """

    if os.path.exists(data_dir) is False:
        os.makedirs(data_dir)

    #     retrieve Gaia DR data by submitting list of source_id to GACS
    output_file = os.path.join(data_dir, '{}_{}_sources.parquet'.format(gaia_data_release, analysis_dataset_name))

    if (not os.path.isfile(output_file)) or (overwrite_query):


        gaia = Gaia
        if shared_user_name is not None:
            if getattr(gaia, '_TapPlus__isLoggedIn') is False:
                gaia.login()
            table_name = 'user_{}'.format(shared_user_name)
        else:
            table_name = '{}'.format(gaia_data_release)

        if source_id_array is not None:
            input_table_name = '{}_source_id'.format(analysis_dataset_name)
            input_table = os.path.join(data_dir, '%s.vot' % input_table_name)
            source_id_column_name = 'source_id'
            Table([source_id_array], names=['source_id']).write(input_table, format='votable', overwrite=True)

            query = '''
            SELECT * FROM
            (select * FROM tap_upload.{0}) AS input
            INNER JOIN
            (select * FROM {1}.{3}) AS gdr
            ON (input.{2} = gdr.source_id)
            ;'''.format(input_table_name, table_name, source_id_column_name, gaia_table_name)

            job = gaia.launch_job_async(query=query, upload_resource=input_table,
                                  upload_table_name=input_table_name, verbose=True)
                                        # dump_to_file=True, output_file=output_file)
        else:
            query = 'select * FROM {}.{};'.format(table_name, gaia_table_name)
            job = gaia.launch_job_async(query=query, verbose=True)

        table = job.get_results()

        df = table.to_pandas()
        df.to_parquet(output_file)
    else:
        df = pd.read_parquet(output_file)

    print('Retrieved {} rows from {}.{}'.format(len(df), gaia_data_release, gaia_table_name))

    return df
